# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from .character_icon import CharacterIcon
from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class OnlinePlayer:
    """
    A player in the online list
    """
    _byte_size: int = 0
    _name: str
    _title: str
    _level: int
    _icon: CharacterIcon
    _class_id: int
    _guild_tag: str

    def __init__(self, *, name: str, title: str, level: int, icon: CharacterIcon, class_id: int, guild_tag: str):
        """
        Create a new instance of OnlinePlayer.

        Args:
            name (str): 
            title (str): 
            level (int): (Value range is 0-252.)
            icon (CharacterIcon): 
            class_id (int): (Value range is 0-252.)
            guild_tag (str): 
        """
        self._name = name
        self._title = title
        self._level = level
        self._icon = icon
        self._class_id = class_id
        self._guild_tag = guild_tag

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def name(self) -> str:
        return self._name

    @property
    def title(self) -> str:
        return self._title

    @property
    def level(self) -> int:
        return self._level

    @property
    def icon(self) -> CharacterIcon:
        return self._icon

    @property
    def class_id(self) -> int:
        return self._class_id

    @property
    def guild_tag(self) -> str:
        return self._guild_tag

    @staticmethod
    def serialize(writer: EoWriter, data: "OnlinePlayer") -> None:
        """
        Serializes an instance of `OnlinePlayer` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (OnlinePlayer): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.string_sanitization_mode = True
            if data._name is None:
                raise SerializationError("name must be provided.")
            writer.add_string(data._name)
            writer.add_byte(0xFF)
            if data._title is None:
                raise SerializationError("title must be provided.")
            writer.add_string(data._title)
            writer.add_byte(0xFF)
            if data._level is None:
                raise SerializationError("level must be provided.")
            writer.add_char(data._level)
            if data._icon is None:
                raise SerializationError("icon must be provided.")
            writer.add_char(int(data._icon))
            if data._class_id is None:
                raise SerializationError("class_id must be provided.")
            writer.add_char(data._class_id)
            if data._guild_tag is None:
                raise SerializationError("guild_tag must be provided.")
            writer.add_string(data._guild_tag)
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "OnlinePlayer":
        """
        Deserializes an instance of `OnlinePlayer` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            OnlinePlayer: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.chunked_reading_mode = True
            name = reader.get_string()
            reader.next_chunk()
            title = reader.get_string()
            reader.next_chunk()
            level = reader.get_char()
            icon = CharacterIcon(reader.get_char())
            class_id = reader.get_char()
            guild_tag = reader.get_string()
            reader.chunked_reading_mode = False
            result = OnlinePlayer(name=name, title=title, level=level, icon=icon, class_id=class_id, guild_tag=guild_tag)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"OnlinePlayer(byte_size={repr(self._byte_size)}, name={repr(self._name)}, title={repr(self._title)}, level={repr(self._level)}, icon={repr(self._icon)}, class_id={repr(self._class_id)}, guild_tag={repr(self._guild_tag)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ...protocol_enum_meta import ProtocolEnumMeta

class MapMusicControl(IntEnum, metaclass=ProtocolEnumMeta):
    """
    How background music should be played on a map
    """
    InterruptIfDifferentPlayOnce = 0
    InterruptPlayOnce = 1
    FinishPlayOnce = 2
    InterruptIfDifferentPlayRepeat = 3
    InterruptPlayRepeat = 4
    FinishPlayRepeat = 5
    InterruptPlayNothing = 6



def encode_string(bytes: bytearray) -> None:
    """
    Encodes a string by inverting the bytes and then reversing them.

    This is an in-place operation.

    Args:
        bytes (bytearray): The byte array to encode.
    """
    _invert_characters(bytes)
    bytes.reverse()


def decode_string(bytes: bytearray) -> None:
    """
    Decodes a string by reversing the bytes and then inverting them.

    This is an in-place operation.

    Args:
        bytes (bytearray): The byte array to decode.
    """
    bytes.reverse()
    _invert_characters(bytes)


def _invert_characters(bytes: bytearray) -> None:
    """
    Inverts characters in a byte array.

    This is an in-place operation.

    Args:
        bytes (bytearray): The byte array to invert.
    """
    flippy = len(bytes) % 2 == 1

    for i in range(len(bytes)):
        c = bytes[i]
        f = 0

        if flippy:
            f = 0x2E
            if c >= 0x50:
                f *= -1

        if 0x22 <= c <= 0x7E:
            bytes[i] = 0x9F - c - f

        flippy = not flippy



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from .trade_item_data import TradeItemData
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class TradeUseServerPacket(Packet):
    """
    Trade completed
    """
    _byte_size: int = 0
    _trade_data: tuple[TradeItemData, ...]

    def __init__(self, *, trade_data: Iterable[TradeItemData]):
        """
        Create a new instance of TradeUseServerPacket.

        Args:
            trade_data (Iterable[TradeItemData]): (Length must be `2`.)
        """
        self._trade_data = tuple(trade_data)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def trade_data(self) -> tuple[TradeItemData, ...]:
        return self._trade_data

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Trade

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Use

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        TradeUseServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "TradeUseServerPacket") -> None:
        """
        Serializes an instance of `TradeUseServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (TradeUseServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.string_sanitization_mode = True
            if data._trade_data is None:
                raise SerializationError("trade_data must be provided.")
            if len(data._trade_data) != 2:
                raise SerializationError(f"Expected length of trade_data to be exactly 2, got {len(data._trade_data)}.")
            for i in range(2):
                TradeItemData.serialize(writer, data._trade_data[i])
                writer.add_byte(0xFF)
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "TradeUseServerPacket":
        """
        Deserializes an instance of `TradeUseServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            TradeUseServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.chunked_reading_mode = True
            trade_data = []
            for i in range(2):
                trade_data.append(TradeItemData.deserialize(reader))
                reader.next_chunk()
            reader.chunked_reading_mode = False
            result = TradeUseServerPacket(trade_data=trade_data)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"TradeUseServerPacket(byte_size={repr(self._byte_size)}, trade_data={repr(self._trade_data)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ...direction import Direction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class AvatarReplyServerPacket(Packet):
    """
    Nearby player hit by another player
    """
    _byte_size: int = 0
    _player_id: int
    _victim_id: int
    _damage: int
    _direction: Direction
    _hp_percentage: int
    _dead: bool

    def __init__(self, *, player_id: int, victim_id: int, damage: int, direction: Direction, hp_percentage: int, dead: bool):
        """
        Create a new instance of AvatarReplyServerPacket.

        Args:
            player_id (int): (Value range is 0-64008.)
            victim_id (int): (Value range is 0-64008.)
            damage (int): (Value range is 0-16194276.)
            direction (Direction): 
            hp_percentage (int): (Value range is 0-252.)
            dead (bool): 
        """
        self._player_id = player_id
        self._victim_id = victim_id
        self._damage = damage
        self._direction = direction
        self._hp_percentage = hp_percentage
        self._dead = dead

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def player_id(self) -> int:
        return self._player_id

    @property
    def victim_id(self) -> int:
        return self._victim_id

    @property
    def damage(self) -> int:
        return self._damage

    @property
    def direction(self) -> Direction:
        return self._direction

    @property
    def hp_percentage(self) -> int:
        return self._hp_percentage

    @property
    def dead(self) -> bool:
        return self._dead

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Avatar

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Reply

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        AvatarReplyServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "AvatarReplyServerPacket") -> None:
        """
        Serializes an instance of `AvatarReplyServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (AvatarReplyServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._player_id is None:
                raise SerializationError("player_id must be provided.")
            writer.add_short(data._player_id)
            if data._victim_id is None:
                raise SerializationError("victim_id must be provided.")
            writer.add_short(data._victim_id)
            if data._damage is None:
                raise SerializationError("damage must be provided.")
            writer.add_three(data._damage)
            if data._direction is None:
                raise SerializationError("direction must be provided.")
            writer.add_char(int(data._direction))
            if data._hp_percentage is None:
                raise SerializationError("hp_percentage must be provided.")
            writer.add_char(data._hp_percentage)
            if data._dead is None:
                raise SerializationError("dead must be provided.")
            writer.add_char(1 if data._dead else 0)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "AvatarReplyServerPacket":
        """
        Deserializes an instance of `AvatarReplyServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            AvatarReplyServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            player_id = reader.get_short()
            victim_id = reader.get_short()
            damage = reader.get_three()
            direction = Direction(reader.get_char())
            hp_percentage = reader.get_char()
            dead = reader.get_char() != 0
            result = AvatarReplyServerPacket(player_id=player_id, victim_id=victim_id, damage=damage, direction=direction, hp_percentage=hp_percentage, dead=dead)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"AvatarReplyServerPacket(byte_size={repr(self._byte_size)}, player_id={repr(self._player_id)}, victim_id={repr(self._victim_id)}, damage={repr(self._damage)}, direction={repr(self._direction)}, hp_percentage={repr(self._hp_percentage)}, dead={repr(self._dead)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from typing import Union
from .train_type import TrainType
from .stat_id import StatId
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class StatSkillAddClientPacket(Packet):
    """
    Spending a stat point on a stat or skill
    """
    _byte_size: int = 0
    _action_type: TrainType
    _action_type_data: 'StatSkillAddClientPacket.ActionTypeData'

    def __init__(self, *, action_type: TrainType, action_type_data: 'StatSkillAddClientPacket.ActionTypeData' = None):
        """
        Create a new instance of StatSkillAddClientPacket.

        Args:
            action_type (TrainType): 
            action_type_data (StatSkillAddClientPacket.ActionTypeData): Data associated with the `action_type` field.
        """
        self._action_type = action_type
        self._action_type_data = action_type_data

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def action_type(self) -> TrainType:
        return self._action_type

    @property
    def action_type_data(self) -> 'StatSkillAddClientPacket.ActionTypeData':
        """
        StatSkillAddClientPacket.ActionTypeData: Data associated with the `action_type` field.
        """
        return self._action_type_data

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.StatSkill

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Add

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        StatSkillAddClientPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "StatSkillAddClientPacket") -> None:
        """
        Serializes an instance of `StatSkillAddClientPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (StatSkillAddClientPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._action_type is None:
                raise SerializationError("action_type must be provided.")
            writer.add_char(int(data._action_type))
            if data._action_type == TrainType.Stat:
                if not isinstance(data._action_type_data, StatSkillAddClientPacket.ActionTypeDataStat):
                    raise SerializationError("Expected action_type_data to be type StatSkillAddClientPacket.ActionTypeDataStat for action_type " + TrainType(data._action_type).name + ".")
                StatSkillAddClientPacket.ActionTypeDataStat.serialize(writer, data._action_type_data)
            elif data._action_type == TrainType.Skill:
                if not isinstance(data._action_type_data, StatSkillAddClientPacket.ActionTypeDataSkill):
                    raise SerializationError("Expected action_type_data to be type StatSkillAddClientPacket.ActionTypeDataSkill for action_type " + TrainType(data._action_type).name + ".")
                StatSkillAddClientPacket.ActionTypeDataSkill.serialize(writer, data._action_type_data)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "StatSkillAddClientPacket":
        """
        Deserializes an instance of `StatSkillAddClientPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            StatSkillAddClientPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            action_type = TrainType(reader.get_char())
            action_type_data: StatSkillAddClientPacket.ActionTypeData = None
            if action_type == TrainType.Stat:
                action_type_data = StatSkillAddClientPacket.ActionTypeDataStat.deserialize(reader)
            elif action_type == TrainType.Skill:
                action_type_data = StatSkillAddClientPacket.ActionTypeDataSkill.deserialize(reader)
            result = StatSkillAddClientPacket(action_type=action_type, action_type_data=action_type_data)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"StatSkillAddClientPacket(byte_size={repr(self._byte_size)}, action_type={repr(self._action_type)}, action_type_data={repr(self._action_type_data)})"

    ActionTypeData = Union['StatSkillAddClientPacket.ActionTypeDataStat', 'StatSkillAddClientPacket.ActionTypeDataSkill', None]
    """
    Data associated with different values of the `action_type` field.
    """

    class ActionTypeDataStat:
        """
        Data associated with action_type value TrainType.Stat
        """
        _byte_size: int = 0
        _stat_id: StatId

        def __init__(self, *, stat_id: StatId):
            """
            Create a new instance of StatSkillAddClientPacket.ActionTypeDataStat.

            Args:
                stat_id (StatId): 
            """
            self._stat_id = stat_id

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def stat_id(self) -> StatId:
            return self._stat_id

        @staticmethod
        def serialize(writer: EoWriter, data: "StatSkillAddClientPacket.ActionTypeDataStat") -> None:
            """
            Serializes an instance of `StatSkillAddClientPacket.ActionTypeDataStat` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (StatSkillAddClientPacket.ActionTypeDataStat): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._stat_id is None:
                    raise SerializationError("stat_id must be provided.")
                writer.add_short(int(data._stat_id))
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "StatSkillAddClientPacket.ActionTypeDataStat":
            """
            Deserializes an instance of `StatSkillAddClientPacket.ActionTypeDataStat` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                StatSkillAddClientPacket.ActionTypeDataStat: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                stat_id = StatId(reader.get_short())
                result = StatSkillAddClientPacket.ActionTypeDataStat(stat_id=stat_id)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"StatSkillAddClientPacket.ActionTypeDataStat(byte_size={repr(self._byte_size)}, stat_id={repr(self._stat_id)})"

    class ActionTypeDataSkill:
        """
        Data associated with action_type value TrainType.Skill
        """
        _byte_size: int = 0
        _spell_id: int

        def __init__(self, *, spell_id: int):
            """
            Create a new instance of StatSkillAddClientPacket.ActionTypeDataSkill.

            Args:
                spell_id (int): (Value range is 0-64008.)
            """
            self._spell_id = spell_id

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def spell_id(self) -> int:
            return self._spell_id

        @staticmethod
        def serialize(writer: EoWriter, data: "StatSkillAddClientPacket.ActionTypeDataSkill") -> None:
            """
            Serializes an instance of `StatSkillAddClientPacket.ActionTypeDataSkill` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (StatSkillAddClientPacket.ActionTypeDataSkill): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._spell_id is None:
                    raise SerializationError("spell_id must be provided.")
                writer.add_short(data._spell_id)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "StatSkillAddClientPacket.ActionTypeDataSkill":
            """
            Deserializes an instance of `StatSkillAddClientPacket.ActionTypeDataSkill` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                StatSkillAddClientPacket.ActionTypeDataSkill: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                spell_id = reader.get_short()
                result = StatSkillAddClientPacket.ActionTypeDataSkill(spell_id=spell_id)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"StatSkillAddClientPacket.ActionTypeDataSkill(byte_size={repr(self._byte_size)}, spell_id={repr(self._spell_id)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ...protocol_enum_meta import ProtocolEnumMeta

class ItemType(IntEnum, metaclass=ProtocolEnumMeta):
    General = 0
    Reserved1 = 1
    Currency = 2
    Heal = 3
    Teleport = 4
    Reserved5 = 5
    ExpReward = 6
    Reserved7 = 7
    Reserved8 = 8
    Key = 9
    Weapon = 10
    Shield = 11
    Armor = 12
    Hat = 13
    Boots = 14
    Gloves = 15
    Accessory = 16
    Belt = 17
    Necklace = 18
    Ring = 19
    Armlet = 20
    Bracer = 21
    Alcohol = 22
    EffectPotion = 23
    HairDye = 24
    CureCurse = 25
    Reserved26 = 26
    Reserved27 = 27
    Reserved28 = 28
    Reserved29 = 29



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class InitBanType(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Ban type sent with INIT_INIT packet.
    The official client treats a value &gt;= 2 as Permanent. Otherwise, it's Temporary.
    """
    Temporary = 1
    Permanent = 2



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from typing import Union
from .map_damage_type import MapDamageType
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class EffectSpecServerPacket(Packet):
    """
    Taking spike or tp drain damage
    """
    _byte_size: int = 0
    _map_damage_type: MapDamageType
    _map_damage_type_data: 'EffectSpecServerPacket.MapDamageTypeData'

    def __init__(self, *, map_damage_type: MapDamageType, map_damage_type_data: 'EffectSpecServerPacket.MapDamageTypeData' = None):
        """
        Create a new instance of EffectSpecServerPacket.

        Args:
            map_damage_type (MapDamageType): 
            map_damage_type_data (EffectSpecServerPacket.MapDamageTypeData): Data associated with the `map_damage_type` field.
        """
        self._map_damage_type = map_damage_type
        self._map_damage_type_data = map_damage_type_data

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def map_damage_type(self) -> MapDamageType:
        return self._map_damage_type

    @property
    def map_damage_type_data(self) -> 'EffectSpecServerPacket.MapDamageTypeData':
        """
        EffectSpecServerPacket.MapDamageTypeData: Data associated with the `map_damage_type` field.
        """
        return self._map_damage_type_data

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Effect

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Spec

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        EffectSpecServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "EffectSpecServerPacket") -> None:
        """
        Serializes an instance of `EffectSpecServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (EffectSpecServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._map_damage_type is None:
                raise SerializationError("map_damage_type must be provided.")
            writer.add_char(int(data._map_damage_type))
            if data._map_damage_type == MapDamageType.TpDrain:
                if not isinstance(data._map_damage_type_data, EffectSpecServerPacket.MapDamageTypeDataTpDrain):
                    raise SerializationError("Expected map_damage_type_data to be type EffectSpecServerPacket.MapDamageTypeDataTpDrain for map_damage_type " + MapDamageType(data._map_damage_type).name + ".")
                EffectSpecServerPacket.MapDamageTypeDataTpDrain.serialize(writer, data._map_damage_type_data)
            elif data._map_damage_type == MapDamageType.Spikes:
                if not isinstance(data._map_damage_type_data, EffectSpecServerPacket.MapDamageTypeDataSpikes):
                    raise SerializationError("Expected map_damage_type_data to be type EffectSpecServerPacket.MapDamageTypeDataSpikes for map_damage_type " + MapDamageType(data._map_damage_type).name + ".")
                EffectSpecServerPacket.MapDamageTypeDataSpikes.serialize(writer, data._map_damage_type_data)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "EffectSpecServerPacket":
        """
        Deserializes an instance of `EffectSpecServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            EffectSpecServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            map_damage_type = MapDamageType(reader.get_char())
            map_damage_type_data: EffectSpecServerPacket.MapDamageTypeData = None
            if map_damage_type == MapDamageType.TpDrain:
                map_damage_type_data = EffectSpecServerPacket.MapDamageTypeDataTpDrain.deserialize(reader)
            elif map_damage_type == MapDamageType.Spikes:
                map_damage_type_data = EffectSpecServerPacket.MapDamageTypeDataSpikes.deserialize(reader)
            result = EffectSpecServerPacket(map_damage_type=map_damage_type, map_damage_type_data=map_damage_type_data)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"EffectSpecServerPacket(byte_size={repr(self._byte_size)}, map_damage_type={repr(self._map_damage_type)}, map_damage_type_data={repr(self._map_damage_type_data)})"

    MapDamageTypeData = Union['EffectSpecServerPacket.MapDamageTypeDataTpDrain', 'EffectSpecServerPacket.MapDamageTypeDataSpikes', None]
    """
    Data associated with different values of the `map_damage_type` field.
    """

    class MapDamageTypeDataTpDrain:
        """
        Data associated with map_damage_type value MapDamageType.TpDrain
        """
        _byte_size: int = 0
        _tp_damage: int
        _tp: int
        _max_tp: int

        def __init__(self, *, tp_damage: int, tp: int, max_tp: int):
            """
            Create a new instance of EffectSpecServerPacket.MapDamageTypeDataTpDrain.

            Args:
                tp_damage (int): (Value range is 0-64008.)
                tp (int): (Value range is 0-64008.)
                max_tp (int): (Value range is 0-64008.)
            """
            self._tp_damage = tp_damage
            self._tp = tp
            self._max_tp = max_tp

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def tp_damage(self) -> int:
            return self._tp_damage

        @property
        def tp(self) -> int:
            return self._tp

        @property
        def max_tp(self) -> int:
            return self._max_tp

        @staticmethod
        def serialize(writer: EoWriter, data: "EffectSpecServerPacket.MapDamageTypeDataTpDrain") -> None:
            """
            Serializes an instance of `EffectSpecServerPacket.MapDamageTypeDataTpDrain` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (EffectSpecServerPacket.MapDamageTypeDataTpDrain): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._tp_damage is None:
                    raise SerializationError("tp_damage must be provided.")
                writer.add_short(data._tp_damage)
                if data._tp is None:
                    raise SerializationError("tp must be provided.")
                writer.add_short(data._tp)
                if data._max_tp is None:
                    raise SerializationError("max_tp must be provided.")
                writer.add_short(data._max_tp)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "EffectSpecServerPacket.MapDamageTypeDataTpDrain":
            """
            Deserializes an instance of `EffectSpecServerPacket.MapDamageTypeDataTpDrain` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                EffectSpecServerPacket.MapDamageTypeDataTpDrain: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                tp_damage = reader.get_short()
                tp = reader.get_short()
                max_tp = reader.get_short()
                result = EffectSpecServerPacket.MapDamageTypeDataTpDrain(tp_damage=tp_damage, tp=tp, max_tp=max_tp)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"EffectSpecServerPacket.MapDamageTypeDataTpDrain(byte_size={repr(self._byte_size)}, tp_damage={repr(self._tp_damage)}, tp={repr(self._tp)}, max_tp={repr(self._max_tp)})"

    class MapDamageTypeDataSpikes:
        """
        Data associated with map_damage_type value MapDamageType.Spikes
        """
        _byte_size: int = 0
        _hp_damage: int
        _hp: int
        _max_hp: int

        def __init__(self, *, hp_damage: int, hp: int, max_hp: int):
            """
            Create a new instance of EffectSpecServerPacket.MapDamageTypeDataSpikes.

            Args:
                hp_damage (int): (Value range is 0-64008.)
                hp (int): (Value range is 0-64008.)
                max_hp (int): (Value range is 0-64008.)
            """
            self._hp_damage = hp_damage
            self._hp = hp
            self._max_hp = max_hp

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def hp_damage(self) -> int:
            return self._hp_damage

        @property
        def hp(self) -> int:
            return self._hp

        @property
        def max_hp(self) -> int:
            return self._max_hp

        @staticmethod
        def serialize(writer: EoWriter, data: "EffectSpecServerPacket.MapDamageTypeDataSpikes") -> None:
            """
            Serializes an instance of `EffectSpecServerPacket.MapDamageTypeDataSpikes` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (EffectSpecServerPacket.MapDamageTypeDataSpikes): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._hp_damage is None:
                    raise SerializationError("hp_damage must be provided.")
                writer.add_short(data._hp_damage)
                if data._hp is None:
                    raise SerializationError("hp must be provided.")
                writer.add_short(data._hp)
                if data._max_hp is None:
                    raise SerializationError("max_hp must be provided.")
                writer.add_short(data._max_hp)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "EffectSpecServerPacket.MapDamageTypeDataSpikes":
            """
            Deserializes an instance of `EffectSpecServerPacket.MapDamageTypeDataSpikes` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                EffectSpecServerPacket.MapDamageTypeDataSpikes: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                hp_damage = reader.get_short()
                hp = reader.get_short()
                max_hp = reader.get_short()
                result = EffectSpecServerPacket.MapDamageTypeDataSpikes(hp_damage=hp_damage, hp=hp, max_hp=max_hp)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"EffectSpecServerPacket.MapDamageTypeDataSpikes(byte_size={repr(self._byte_size)}, hp_damage={repr(self._hp_damage)}, hp={repr(self._hp)}, max_hp={repr(self._max_hp)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ...serialization_error import SerializationError
from ....data.eo_writer import EoWriter
from ....data.eo_reader import EoReader

class EcfRecord:
    """
    Record of Class data in an Endless Class File
    """
    _byte_size: int = 0
    _name_length: int
    _name: str
    _parent_type: int
    _stat_group: int
    _str: int
    _intl: int
    _wis: int
    _agi: int
    _con: int
    _cha: int

    def __init__(self, *, name: str, parent_type: int, stat_group: int, str: int, intl: int, wis: int, agi: int, con: int, cha: int):
        """
        Create a new instance of EcfRecord.

        Args:
            name (str): (Length must be 252 or less.)
            parent_type (int): (Value range is 0-252.)
            stat_group (int): (Value range is 0-252.)
            str (int): (Value range is 0-64008.)
            intl (int): (Value range is 0-64008.)
            wis (int): (Value range is 0-64008.)
            agi (int): (Value range is 0-64008.)
            con (int): (Value range is 0-64008.)
            cha (int): (Value range is 0-64008.)
        """
        self._name = name
        self._name_length = len(self._name)
        self._parent_type = parent_type
        self._stat_group = stat_group
        self._str = str
        self._intl = intl
        self._wis = wis
        self._agi = agi
        self._con = con
        self._cha = cha

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def name(self) -> str:
        return self._name

    @property
    def parent_type(self) -> int:
        return self._parent_type

    @property
    def stat_group(self) -> int:
        return self._stat_group

    @property
    def str(self) -> int:
        return self._str

    @property
    def intl(self) -> int:
        return self._intl

    @property
    def wis(self) -> int:
        return self._wis

    @property
    def agi(self) -> int:
        return self._agi

    @property
    def con(self) -> int:
        return self._con

    @property
    def cha(self) -> int:
        return self._cha

    @staticmethod
    def serialize(writer: EoWriter, data: "EcfRecord") -> None:
        """
        Serializes an instance of `EcfRecord` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (EcfRecord): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._name_length is None:
                raise SerializationError("name_length must be provided.")
            writer.add_char(data._name_length)
            if data._name is None:
                raise SerializationError("name must be provided.")
            if len(data._name) > 252:
                raise SerializationError(f"Expected length of name to be 252 or less, got {len(data._name)}.")
            writer.add_fixed_string(data._name, data._name_length, False)
            if data._parent_type is None:
                raise SerializationError("parent_type must be provided.")
            writer.add_char(data._parent_type)
            if data._stat_group is None:
                raise SerializationError("stat_group must be provided.")
            writer.add_char(data._stat_group)
            if data._str is None:
                raise SerializationError("str must be provided.")
            writer.add_short(data._str)
            if data._intl is None:
                raise SerializationError("intl must be provided.")
            writer.add_short(data._intl)
            if data._wis is None:
                raise SerializationError("wis must be provided.")
            writer.add_short(data._wis)
            if data._agi is None:
                raise SerializationError("agi must be provided.")
            writer.add_short(data._agi)
            if data._con is None:
                raise SerializationError("con must be provided.")
            writer.add_short(data._con)
            if data._cha is None:
                raise SerializationError("cha must be provided.")
            writer.add_short(data._cha)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "EcfRecord":
        """
        Deserializes an instance of `EcfRecord` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            EcfRecord: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            name_length = reader.get_char()
            name = reader.get_fixed_string(name_length, False)
            parent_type = reader.get_char()
            stat_group = reader.get_char()
            str = reader.get_short()
            intl = reader.get_short()
            wis = reader.get_short()
            agi = reader.get_short()
            con = reader.get_short()
            cha = reader.get_short()
            result = EcfRecord(name=name, parent_type=parent_type, stat_group=stat_group, str=str, intl=intl, wis=wis, agi=agi, con=con, cha=cha)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"EcfRecord(byte_size={repr(self._byte_size)}, name={repr(self._name)}, parent_type={repr(self._parent_type)}, stat_group={repr(self._stat_group)}, str={repr(self._str)}, intl={repr(self._intl)}, wis={repr(self._wis)}, agi={repr(self._agi)}, con={repr(self._con)}, cha={repr(self._cha)})"



import pytest
from collections import namedtuple
from eolib.data.string_encoding_utils import encode_string, decode_string

TCase = namedtuple("TCase", ["decoded", "encoded"])

TEST_CASES = [
    TCase("Hello, World!", "!;a-^H s^3a:)"),
    TCase(
        "We're ¼ of the way there, so ¾ is remaining.",
        "C8_6_6l2h- ,d ¾ ^, sh-h7Y T>V h7Y g0 ¼ :[xhH",
    ),
    TCase("64² = 4096", ";fAk b ²=i"),
    TCase("© FÒÖ BÃR BÅZ 2014", "=nAm EÅ] MÃ] ÖÒY ©"),
    TCase('Öxxö Xööx "Lëïth Säë" - "Ÿ"', "OŸO D OëäL 7YïëSO UööG öU'Ö"),
    TCase("Padded with 0xFFÿÿÿÿÿÿÿÿ", "ÿÿÿÿÿÿÿÿ+YUo 7Y6V i:i;lO"),
]


def idfn(test_case: TCase) -> str:
    return str(test_case.decoded)


@pytest.mark.parametrize("test_case", TEST_CASES, ids=idfn)
def test_encode_string(test_case: TCase) -> None:
    decoded_bytes = bytearray(test_case.decoded, "windows-1252")

    encode_string(decoded_bytes)
    encoded = decoded_bytes.decode("windows-1252")

    assert encoded == test_case.encoded


@pytest.mark.parametrize("test_case", TEST_CASES, ids=idfn)
def test_decode_string(test_case: TCase) -> None:
    encoded_bytes = bytearray(test_case.encoded, "windows-1252")

    decode_string(encoded_bytes)
    decoded = encoded_bytes.decode("windows-1252")

    assert decoded == test_case.decoded



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from .character_stats_info_lookup import CharacterStatsInfoLookup
from .big_coords import BigCoords
from ..weight import Weight
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class AdminInteractTellServerPacket(Packet):
    """
    Admin character info lookup
    """
    _byte_size: int = 0
    _name: str
    _usage: int
    _gold_bank: int
    _exp: int
    _level: int
    _map_id: int
    _map_coords: BigCoords
    _stats: CharacterStatsInfoLookup
    _weight: Weight

    def __init__(self, *, name: str, usage: int, gold_bank: int, exp: int, level: int, map_id: int, map_coords: BigCoords, stats: CharacterStatsInfoLookup, weight: Weight):
        """
        Create a new instance of AdminInteractTellServerPacket.

        Args:
            name (str): 
            usage (int): (Value range is 0-4097152080.)
            gold_bank (int): (Value range is 0-4097152080.)
            exp (int): (Value range is 0-4097152080.)
            level (int): (Value range is 0-252.)
            map_id (int): (Value range is 0-64008.)
            map_coords (BigCoords): 
            stats (CharacterStatsInfoLookup): 
            weight (Weight): 
        """
        self._name = name
        self._usage = usage
        self._gold_bank = gold_bank
        self._exp = exp
        self._level = level
        self._map_id = map_id
        self._map_coords = map_coords
        self._stats = stats
        self._weight = weight

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def name(self) -> str:
        return self._name

    @property
    def usage(self) -> int:
        return self._usage

    @property
    def gold_bank(self) -> int:
        return self._gold_bank

    @property
    def exp(self) -> int:
        return self._exp

    @property
    def level(self) -> int:
        return self._level

    @property
    def map_id(self) -> int:
        return self._map_id

    @property
    def map_coords(self) -> BigCoords:
        return self._map_coords

    @property
    def stats(self) -> CharacterStatsInfoLookup:
        return self._stats

    @property
    def weight(self) -> Weight:
        return self._weight

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.AdminInteract

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Tell

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        AdminInteractTellServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "AdminInteractTellServerPacket") -> None:
        """
        Serializes an instance of `AdminInteractTellServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (AdminInteractTellServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.string_sanitization_mode = True
            if data._name is None:
                raise SerializationError("name must be provided.")
            writer.add_string(data._name)
            writer.add_byte(0xFF)
            if data._usage is None:
                raise SerializationError("usage must be provided.")
            writer.add_int(data._usage)
            writer.add_byte(0xFF)
            if data._gold_bank is None:
                raise SerializationError("gold_bank must be provided.")
            writer.add_int(data._gold_bank)
            writer.add_byte(0xFF)
            if data._exp is None:
                raise SerializationError("exp must be provided.")
            writer.add_int(data._exp)
            if data._level is None:
                raise SerializationError("level must be provided.")
            writer.add_char(data._level)
            if data._map_id is None:
                raise SerializationError("map_id must be provided.")
            writer.add_short(data._map_id)
            if data._map_coords is None:
                raise SerializationError("map_coords must be provided.")
            BigCoords.serialize(writer, data._map_coords)
            if data._stats is None:
                raise SerializationError("stats must be provided.")
            CharacterStatsInfoLookup.serialize(writer, data._stats)
            if data._weight is None:
                raise SerializationError("weight must be provided.")
            Weight.serialize(writer, data._weight)
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "AdminInteractTellServerPacket":
        """
        Deserializes an instance of `AdminInteractTellServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            AdminInteractTellServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.chunked_reading_mode = True
            name = reader.get_string()
            reader.next_chunk()
            usage = reader.get_int()
            reader.next_chunk()
            gold_bank = reader.get_int()
            reader.next_chunk()
            exp = reader.get_int()
            level = reader.get_char()
            map_id = reader.get_short()
            map_coords = BigCoords.deserialize(reader)
            stats = CharacterStatsInfoLookup.deserialize(reader)
            weight = Weight.deserialize(reader)
            reader.chunked_reading_mode = False
            result = AdminInteractTellServerPacket(name=name, usage=usage, gold_bank=gold_bank, exp=exp, level=level, map_id=map_id, map_coords=map_coords, stats=stats, weight=weight)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"AdminInteractTellServerPacket(byte_size={repr(self._byte_size)}, name={repr(self._name)}, usage={repr(self._usage)}, gold_bank={repr(self._gold_bank)}, exp={repr(self._exp)}, level={repr(self._level)}, map_id={repr(self._map_id)}, map_coords={repr(self._map_coords)}, stats={repr(self._stats)}, weight={repr(self._weight)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ...protocol_enum_meta import ProtocolEnumMeta

class PacketFamily(IntEnum, metaclass=ProtocolEnumMeta):
    """
    The type of operation that a packet performs.
    Part of the unique packet ID.
    """
    Connection = 1
    Account = 2
    Character = 3
    Login = 4
    Welcome = 5
    Walk = 6
    Face = 7
    Chair = 8
    Emote = 9
    Attack = 11
    Spell = 12
    Shop = 13
    Item = 14
    StatSkill = 16
    Global = 17
    Talk = 18
    Warp = 19
    Jukebox = 21
    Players = 22
    Avatar = 23
    Party = 24
    Refresh = 25
    Npc = 26
    PlayerRange = 27
    NpcRange = 28
    Range = 29
    Paperdoll = 30
    Effect = 31
    Trade = 32
    Chest = 33
    Door = 34
    Message = 35
    Bank = 36
    Locker = 37
    Barber = 38
    Guild = 39
    Music = 40
    Sit = 41
    Recover = 42
    Board = 43
    Cast = 44
    Arena = 45
    Priest = 46
    Marriage = 47
    AdminInteract = 48
    Citizen = 49
    Quest = 50
    Book = 51
    Error = 250
    Init = 255



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from .inn_question_record import InnQuestionRecord
from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class InnRecord:
    """
    Record of Inn data in an Endless Inn File
    """
    _byte_size: int = 0
    _behavior_id: int
    _name_length: int
    _name: str
    _spawn_map: int
    _spawn_x: int
    _spawn_y: int
    _sleep_map: int
    _sleep_x: int
    _sleep_y: int
    _alternate_spawn_enabled: bool
    _alternate_spawn_map: int
    _alternate_spawn_x: int
    _alternate_spawn_y: int
    _questions: tuple[InnQuestionRecord, ...]

    def __init__(self, *, behavior_id: int, name: str, spawn_map: int, spawn_x: int, spawn_y: int, sleep_map: int, sleep_x: int, sleep_y: int, alternate_spawn_enabled: bool, alternate_spawn_map: int, alternate_spawn_x: int, alternate_spawn_y: int, questions: Iterable[InnQuestionRecord]):
        """
        Create a new instance of InnRecord.

        Args:
            behavior_id (int): Behavior ID of the NPC that runs the inn. 0 for default inn (Value range is 0-64008.)
            name (str): (Length must be 252 or less.)
            spawn_map (int): ID of the map the player is sent to after respawning (Value range is 0-64008.)
            spawn_x (int): X coordinate of the map the player is sent to after respawning (Value range is 0-252.)
            spawn_y (int): Y coordinate of the map the player is sent to after respawning (Value range is 0-252.)
            sleep_map (int): ID of the map the player is sent to after sleeping at the inn (Value range is 0-64008.)
            sleep_x (int): X coordinate of the map the player is sent to after sleeping at the inn (Value range is 0-252.)
            sleep_y (int): Y coordinate of the map the player is sent to after sleeping at the inn (Value range is 0-252.)
            alternate_spawn_enabled (bool): Flag for an alternate spawn point. If true, the server will use this alternate spawn map, x, and, y based on some other condition.  In the official server, this is used to respawn new characters on the noob island until they reach a certain level.
            alternate_spawn_map (int): (Value range is 0-64008.)
            alternate_spawn_x (int): (Value range is 0-252.)
            alternate_spawn_y (int): (Value range is 0-252.)
            questions (Iterable[InnQuestionRecord]): (Length must be `3`.)
        """
        self._behavior_id = behavior_id
        self._name = name
        self._name_length = len(self._name)
        self._spawn_map = spawn_map
        self._spawn_x = spawn_x
        self._spawn_y = spawn_y
        self._sleep_map = sleep_map
        self._sleep_x = sleep_x
        self._sleep_y = sleep_y
        self._alternate_spawn_enabled = alternate_spawn_enabled
        self._alternate_spawn_map = alternate_spawn_map
        self._alternate_spawn_x = alternate_spawn_x
        self._alternate_spawn_y = alternate_spawn_y
        self._questions = tuple(questions)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def behavior_id(self) -> int:
        """
        Behavior ID of the NPC that runs the inn. 0 for default inn
        """
        return self._behavior_id

    @property
    def name(self) -> str:
        return self._name

    @property
    def spawn_map(self) -> int:
        """
        ID of the map the player is sent to after respawning
        """
        return self._spawn_map

    @property
    def spawn_x(self) -> int:
        """
        X coordinate of the map the player is sent to after respawning
        """
        return self._spawn_x

    @property
    def spawn_y(self) -> int:
        """
        Y coordinate of the map the player is sent to after respawning
        """
        return self._spawn_y

    @property
    def sleep_map(self) -> int:
        """
        ID of the map the player is sent to after sleeping at the inn
        """
        return self._sleep_map

    @property
    def sleep_x(self) -> int:
        """
        X coordinate of the map the player is sent to after sleeping at the inn
        """
        return self._sleep_x

    @property
    def sleep_y(self) -> int:
        """
        Y coordinate of the map the player is sent to after sleeping at the inn
        """
        return self._sleep_y

    @property
    def alternate_spawn_enabled(self) -> bool:
        """
        Flag for an alternate spawn point. If true, the server will use this alternate spawn
        map, x, and, y based on some other condition.

        In the official server, this is used to respawn new characters on the noob island
        until they reach a certain level.
        """
        return self._alternate_spawn_enabled

    @property
    def alternate_spawn_map(self) -> int:
        return self._alternate_spawn_map

    @property
    def alternate_spawn_x(self) -> int:
        return self._alternate_spawn_x

    @property
    def alternate_spawn_y(self) -> int:
        return self._alternate_spawn_y

    @property
    def questions(self) -> tuple[InnQuestionRecord, ...]:
        return self._questions

    @staticmethod
    def serialize(writer: EoWriter, data: "InnRecord") -> None:
        """
        Serializes an instance of `InnRecord` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (InnRecord): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._behavior_id is None:
                raise SerializationError("behavior_id must be provided.")
            writer.add_short(data._behavior_id)
            if data._name_length is None:
                raise SerializationError("name_length must be provided.")
            writer.add_char(data._name_length)
            if data._name is None:
                raise SerializationError("name must be provided.")
            if len(data._name) > 252:
                raise SerializationError(f"Expected length of name to be 252 or less, got {len(data._name)}.")
            writer.add_fixed_string(data._name, data._name_length, False)
            if data._spawn_map is None:
                raise SerializationError("spawn_map must be provided.")
            writer.add_short(data._spawn_map)
            if data._spawn_x is None:
                raise SerializationError("spawn_x must be provided.")
            writer.add_char(data._spawn_x)
            if data._spawn_y is None:
                raise SerializationError("spawn_y must be provided.")
            writer.add_char(data._spawn_y)
            if data._sleep_map is None:
                raise SerializationError("sleep_map must be provided.")
            writer.add_short(data._sleep_map)
            if data._sleep_x is None:
                raise SerializationError("sleep_x must be provided.")
            writer.add_char(data._sleep_x)
            if data._sleep_y is None:
                raise SerializationError("sleep_y must be provided.")
            writer.add_char(data._sleep_y)
            if data._alternate_spawn_enabled is None:
                raise SerializationError("alternate_spawn_enabled must be provided.")
            writer.add_char(1 if data._alternate_spawn_enabled else 0)
            if data._alternate_spawn_map is None:
                raise SerializationError("alternate_spawn_map must be provided.")
            writer.add_short(data._alternate_spawn_map)
            if data._alternate_spawn_x is None:
                raise SerializationError("alternate_spawn_x must be provided.")
            writer.add_char(data._alternate_spawn_x)
            if data._alternate_spawn_y is None:
                raise SerializationError("alternate_spawn_y must be provided.")
            writer.add_char(data._alternate_spawn_y)
            if data._questions is None:
                raise SerializationError("questions must be provided.")
            if len(data._questions) != 3:
                raise SerializationError(f"Expected length of questions to be exactly 3, got {len(data._questions)}.")
            for i in range(3):
                InnQuestionRecord.serialize(writer, data._questions[i])
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "InnRecord":
        """
        Deserializes an instance of `InnRecord` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            InnRecord: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            behavior_id = reader.get_short()
            name_length = reader.get_char()
            name = reader.get_fixed_string(name_length, False)
            spawn_map = reader.get_short()
            spawn_x = reader.get_char()
            spawn_y = reader.get_char()
            sleep_map = reader.get_short()
            sleep_x = reader.get_char()
            sleep_y = reader.get_char()
            alternate_spawn_enabled = reader.get_char() != 0
            alternate_spawn_map = reader.get_short()
            alternate_spawn_x = reader.get_char()
            alternate_spawn_y = reader.get_char()
            questions = []
            for i in range(3):
                questions.append(InnQuestionRecord.deserialize(reader))
            result = InnRecord(behavior_id=behavior_id, name=name, spawn_map=spawn_map, spawn_x=spawn_x, spawn_y=spawn_y, sleep_map=sleep_map, sleep_x=sleep_x, sleep_y=sleep_y, alternate_spawn_enabled=alternate_spawn_enabled, alternate_spawn_map=alternate_spawn_map, alternate_spawn_x=alternate_spawn_x, alternate_spawn_y=alternate_spawn_y, questions=questions)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"InnRecord(byte_size={repr(self._byte_size)}, behavior_id={repr(self._behavior_id)}, name={repr(self._name)}, spawn_map={repr(self._spawn_map)}, spawn_x={repr(self._spawn_x)}, spawn_y={repr(self._spawn_y)}, sleep_map={repr(self._sleep_map)}, sleep_x={repr(self._sleep_x)}, sleep_y={repr(self._sleep_y)}, alternate_spawn_enabled={repr(self._alternate_spawn_enabled)}, alternate_spawn_map={repr(self._alternate_spawn_map)}, alternate_spawn_x={repr(self._alternate_spawn_x)}, alternate_spawn_y={repr(self._alternate_spawn_y)}, questions={repr(self._questions)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class TalkMessageRecord:
    """
    Record of a message that an NPC can say
    """
    _byte_size: int = 0
    _message_length: int
    _message: str

    def __init__(self, *, message: str):
        """
        Create a new instance of TalkMessageRecord.

        Args:
            message (str): (Length must be 252 or less.)
        """
        self._message = message
        self._message_length = len(self._message)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def message(self) -> str:
        return self._message

    @staticmethod
    def serialize(writer: EoWriter, data: "TalkMessageRecord") -> None:
        """
        Serializes an instance of `TalkMessageRecord` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (TalkMessageRecord): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._message_length is None:
                raise SerializationError("message_length must be provided.")
            writer.add_char(data._message_length)
            if data._message is None:
                raise SerializationError("message must be provided.")
            if len(data._message) > 252:
                raise SerializationError(f"Expected length of message to be 252 or less, got {len(data._message)}.")
            writer.add_fixed_string(data._message, data._message_length, False)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "TalkMessageRecord":
        """
        Deserializes an instance of `TalkMessageRecord` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            TalkMessageRecord: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            message_length = reader.get_char()
            message = reader.get_fixed_string(message_length, False)
            result = TalkMessageRecord(message=message)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"TalkMessageRecord(byte_size={repr(self._byte_size)}, message={repr(self._message)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class NpcKillStealProtectionState(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Flag to indicate whether you are able to attack an NPC
    """
    Unprotected = 1
    Protected = 2



import shutil
import os
from argparse import ArgumentParser
from pathlib import Path
from contextlib import suppress
from protocol_code_generator.generate.code_generator import ProtocolCodeGenerator


def clean() -> None:
    generated = _generated_dir()
    if os.path.exists(generated):
        print(f"Removing: {generated}")
        shutil.rmtree(generated)


def generate() -> None:
    code_generator = ProtocolCodeGenerator(_eo_protocol_dir())
    code_generator.generate(_generated_dir())


def _generated_dir() -> Path:
    return Path(__file__).parent.joinpath('src/eolib/protocol/_generated').resolve()


def _eo_protocol_dir() -> Path:
    return Path(__file__).parent.joinpath('eo-protocol/xml').resolve()


if __name__ == "__main__":
    parser = ArgumentParser(description='Helper script for managing generated EO protocol code.')
    parser.add_argument("command", choices=["generate", "clean"])
    args = parser.parse_args()

    clean()

    if args.command != "clean":
        generate()



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class SkillMasterSkillRecord:
    """
    Record of a skill that a Skill Master NPC can teach
    """
    _byte_size: int = 0
    _skill_id: int
    _level_requirement: int
    _class_requirement: int
    _price: int
    _skill_requirements: tuple[int, ...]
    _str_requirement: int
    _int_requirement: int
    _wis_requirement: int
    _agi_requirement: int
    _con_requirement: int
    _cha_requirement: int

    def __init__(self, *, skill_id: int, level_requirement: int, class_requirement: int, price: int, skill_requirements: Iterable[int], str_requirement: int, int_requirement: int, wis_requirement: int, agi_requirement: int, con_requirement: int, cha_requirement: int):
        """
        Create a new instance of SkillMasterSkillRecord.

        Args:
            skill_id (int): (Value range is 0-64008.)
            level_requirement (int): Level required to learn this skill (Value range is 0-252.)
            class_requirement (int): Class required to learn this skill (Value range is 0-252.)
            price (int): (Value range is 0-4097152080.)
            skill_requirements (Iterable[int]): IDs of skills that must be learned before a player can learn this skill (Length must be `4`.) (Element value range is 0-64008.)
            str_requirement (int): Strength required to learn this skill (Value range is 0-64008.)
            int_requirement (int): Intelligence required to learn this skill (Value range is 0-64008.)
            wis_requirement (int): Wisdom required to learn this skill (Value range is 0-64008.)
            agi_requirement (int): Agility required to learn this skill (Value range is 0-64008.)
            con_requirement (int): Constitution required to learn this skill (Value range is 0-64008.)
            cha_requirement (int): Charisma required to learn this skill (Value range is 0-64008.)
        """
        self._skill_id = skill_id
        self._level_requirement = level_requirement
        self._class_requirement = class_requirement
        self._price = price
        self._skill_requirements = tuple(skill_requirements)
        self._str_requirement = str_requirement
        self._int_requirement = int_requirement
        self._wis_requirement = wis_requirement
        self._agi_requirement = agi_requirement
        self._con_requirement = con_requirement
        self._cha_requirement = cha_requirement

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def skill_id(self) -> int:
        return self._skill_id

    @property
    def level_requirement(self) -> int:
        """
        Level required to learn this skill
        """
        return self._level_requirement

    @property
    def class_requirement(self) -> int:
        """
        Class required to learn this skill
        """
        return self._class_requirement

    @property
    def price(self) -> int:
        return self._price

    @property
    def skill_requirements(self) -> tuple[int, ...]:
        """
        IDs of skills that must be learned before a player can learn this skill
        """
        return self._skill_requirements

    @property
    def str_requirement(self) -> int:
        """
        Strength required to learn this skill
        """
        return self._str_requirement

    @property
    def int_requirement(self) -> int:
        """
        Intelligence required to learn this skill
        """
        return self._int_requirement

    @property
    def wis_requirement(self) -> int:
        """
        Wisdom required to learn this skill
        """
        return self._wis_requirement

    @property
    def agi_requirement(self) -> int:
        """
        Agility required to learn this skill
        """
        return self._agi_requirement

    @property
    def con_requirement(self) -> int:
        """
        Constitution required to learn this skill
        """
        return self._con_requirement

    @property
    def cha_requirement(self) -> int:
        """
        Charisma required to learn this skill
        """
        return self._cha_requirement

    @staticmethod
    def serialize(writer: EoWriter, data: "SkillMasterSkillRecord") -> None:
        """
        Serializes an instance of `SkillMasterSkillRecord` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (SkillMasterSkillRecord): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._skill_id is None:
                raise SerializationError("skill_id must be provided.")
            writer.add_short(data._skill_id)
            if data._level_requirement is None:
                raise SerializationError("level_requirement must be provided.")
            writer.add_char(data._level_requirement)
            if data._class_requirement is None:
                raise SerializationError("class_requirement must be provided.")
            writer.add_char(data._class_requirement)
            if data._price is None:
                raise SerializationError("price must be provided.")
            writer.add_int(data._price)
            if data._skill_requirements is None:
                raise SerializationError("skill_requirements must be provided.")
            if len(data._skill_requirements) != 4:
                raise SerializationError(f"Expected length of skill_requirements to be exactly 4, got {len(data._skill_requirements)}.")
            for i in range(4):
                writer.add_short(data._skill_requirements[i])
            if data._str_requirement is None:
                raise SerializationError("str_requirement must be provided.")
            writer.add_short(data._str_requirement)
            if data._int_requirement is None:
                raise SerializationError("int_requirement must be provided.")
            writer.add_short(data._int_requirement)
            if data._wis_requirement is None:
                raise SerializationError("wis_requirement must be provided.")
            writer.add_short(data._wis_requirement)
            if data._agi_requirement is None:
                raise SerializationError("agi_requirement must be provided.")
            writer.add_short(data._agi_requirement)
            if data._con_requirement is None:
                raise SerializationError("con_requirement must be provided.")
            writer.add_short(data._con_requirement)
            if data._cha_requirement is None:
                raise SerializationError("cha_requirement must be provided.")
            writer.add_short(data._cha_requirement)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "SkillMasterSkillRecord":
        """
        Deserializes an instance of `SkillMasterSkillRecord` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            SkillMasterSkillRecord: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            skill_id = reader.get_short()
            level_requirement = reader.get_char()
            class_requirement = reader.get_char()
            price = reader.get_int()
            skill_requirements = []
            for i in range(4):
                skill_requirements.append(reader.get_short())
            str_requirement = reader.get_short()
            int_requirement = reader.get_short()
            wis_requirement = reader.get_short()
            agi_requirement = reader.get_short()
            con_requirement = reader.get_short()
            cha_requirement = reader.get_short()
            result = SkillMasterSkillRecord(skill_id=skill_id, level_requirement=level_requirement, class_requirement=class_requirement, price=price, skill_requirements=skill_requirements, str_requirement=str_requirement, int_requirement=int_requirement, wis_requirement=wis_requirement, agi_requirement=agi_requirement, con_requirement=con_requirement, cha_requirement=cha_requirement)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"SkillMasterSkillRecord(byte_size={repr(self._byte_size)}, skill_id={repr(self._skill_id)}, level_requirement={repr(self._level_requirement)}, class_requirement={repr(self._class_requirement)}, price={repr(self._price)}, skill_requirements={repr(self._skill_requirements)}, str_requirement={repr(self._str_requirement)}, int_requirement={repr(self._int_requirement)}, wis_requirement={repr(self._wis_requirement)}, agi_requirement={repr(self._agi_requirement)}, con_requirement={repr(self._con_requirement)}, cha_requirement={repr(self._cha_requirement)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class SkillMasterReply(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Reply code sent with STATSKILL_REPLY packet.
    Indicates why an action was unsuccessful.
    """
    RemoveItems = 1
    WrongClass = 2



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from .online_player import OnlinePlayer
from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class PlayersList:
    """
    Information about online players
    """
    _byte_size: int = 0
    _players_count: int
    _players: tuple[OnlinePlayer, ...]

    def __init__(self, *, players: Iterable[OnlinePlayer]):
        """
        Create a new instance of PlayersList.

        Args:
            players (Iterable[OnlinePlayer]): (Length must be 64008 or less.)
        """
        self._players = tuple(players)
        self._players_count = len(self._players)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def players(self) -> tuple[OnlinePlayer, ...]:
        return self._players

    @staticmethod
    def serialize(writer: EoWriter, data: "PlayersList") -> None:
        """
        Serializes an instance of `PlayersList` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (PlayersList): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.string_sanitization_mode = True
            if data._players_count is None:
                raise SerializationError("players_count must be provided.")
            writer.add_short(data._players_count)
            writer.add_byte(0xFF)
            if data._players is None:
                raise SerializationError("players must be provided.")
            if len(data._players) > 64008:
                raise SerializationError(f"Expected length of players to be 64008 or less, got {len(data._players)}.")
            for i in range(data._players_count):
                OnlinePlayer.serialize(writer, data._players[i])
                writer.add_byte(0xFF)
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "PlayersList":
        """
        Deserializes an instance of `PlayersList` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            PlayersList: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.chunked_reading_mode = True
            players_count = reader.get_short()
            reader.next_chunk()
            players = []
            for i in range(players_count):
                players.append(OnlinePlayer.deserialize(reader))
                reader.next_chunk()
            reader.chunked_reading_mode = False
            result = PlayersList(players=players)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"PlayersList(byte_size={repr(self._byte_size)}, players={repr(self._players)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class PubFile:
    """
    A pub file (EIF, ENF, ECF, ESF)
    """
    _byte_size: int = 0
    _file_id: int
    _content: bytes

    def __init__(self, *, file_id: int, content: bytes):
        """
        Create a new instance of PubFile.

        Args:
            file_id (int): (Value range is 0-252.)
            content (bytes): 
        """
        self._file_id = file_id
        self._content = content

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def file_id(self) -> int:
        return self._file_id

    @property
    def content(self) -> bytes:
        return self._content

    @staticmethod
    def serialize(writer: EoWriter, data: "PubFile") -> None:
        """
        Serializes an instance of `PubFile` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (PubFile): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._file_id is None:
                raise SerializationError("file_id must be provided.")
            writer.add_char(data._file_id)
            if data._content is None:
                raise SerializationError("content must be provided.")
            writer.add_bytes(data._content)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "PubFile":
        """
        Deserializes an instance of `PubFile` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            PubFile: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            file_id = reader.get_char()
            content = reader.get_bytes(reader.remaining)
            result = PubFile(file_id=file_id, content=content)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"PubFile(byte_size={repr(self._byte_size)}, file_id={repr(self._file_id)}, content={repr(self._content)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ...protocol_enum_meta import ProtocolEnumMeta

class MapTileSpec(IntEnum, metaclass=ProtocolEnumMeta):
    """
    The type of a tile on a map
    """
    Wall = 0
    ChairDown = 1
    ChairLeft = 2
    ChairRight = 3
    ChairUp = 4
    ChairDownRight = 5
    ChairUpLeft = 6
    ChairAll = 7
    Reserved8 = 8
    Chest = 9
    Reserved10 = 10
    Reserved11 = 11
    Reserved12 = 12
    Reserved13 = 13
    Reserved14 = 14
    Reserved15 = 15
    BankVault = 16
    NpcBoundary = 17
    Edge = 18
    FakeWall = 19
    Board1 = 20
    Board2 = 21
    Board3 = 22
    Board4 = 23
    Board5 = 24
    Board6 = 25
    Board7 = 26
    Board8 = 27
    Jukebox = 28
    Jump = 29
    Water = 30
    Reserved31 = 31
    Arena = 32
    AmbientSource = 33
    TimedSpikes = 34
    Spikes = 35
    HiddenSpikes = 36



class SerializationError(Exception):
    """
    Represents an error in serializing a protocol data structure.
    """

    def __init__(self, message: str):
        """
        Constructs a SerializationError with the specified error message.

        Args:
            message (str): The error message.
        """
        super().__init__(message)



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class PlayerKilledState(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Flag to indicate that a player has been killed
    """
    Alive = 1
    Killed = 2



from html import escape
from protocol_code_generator.generate.code_block import CodeBlock
from protocol_code_generator.generate.object_code_generator import FieldData
from protocol_code_generator.type.basic_type import BasicType
from protocol_code_generator.type.blob_type import BlobType
from protocol_code_generator.type.bool_type import BoolType
from protocol_code_generator.type.custom_type import CustomType
from protocol_code_generator.type.enum_type import EnumType
from protocol_code_generator.type.has_underlying_type import HasUnderlyingType
from protocol_code_generator.type.integer_type import IntegerType
from protocol_code_generator.type.length import Length
from protocol_code_generator.type.string_type import StringType
from protocol_code_generator.type.struct_type import StructType
from protocol_code_generator.util.docstring_utils import generate_docstring
from protocol_code_generator.util.number_utils import try_parse_int


class FieldCodeGenerator:
    def __init__(
        self,
        type_factory,
        context,
        data,
        name,
        type_string,
        length_string,
        padded,
        optional,
        hardcoded_value,
        comment,
        array_field,
        delimited,
        trailing_delimiter,
        length_field,
        offset,
    ):
        self._type_factory = type_factory
        self._context = context
        self._data = data
        self._name = name
        self._type_string = type_string
        self._length_string = length_string
        self._padded = padded
        self._optional = optional
        self._hardcoded_value = hardcoded_value
        self._comment = comment
        self._array_field = array_field
        self._delimited = delimited
        self._trailing_delimiter = trailing_delimiter
        self._length_field = length_field
        self._offset = offset
        self._validate()

    def _validate(self):
        self._validate_special_fields()
        self._validate_optional_field()
        self._validate_array_field()
        self._validate_length_field()
        self._validate_unnamed_field()
        self._validate_hardcoded_value()
        self._validate_unique_name()
        self._validate_length_attribute()

    def _validate_special_fields(self):
        if self._array_field and self._length_field:
            raise RuntimeError("A field cannot be both a length field and an array field.")

    def _validate_optional_field(self):
        if not self._optional:
            return

        if self._name is None:
            raise RuntimeError("Optional fields must specify a name.")

    def _validate_array_field(self):
        if self._array_field:
            if self._name is None:
                raise RuntimeError("Array fields must specify a name.")
            if self._hardcoded_value:
                raise RuntimeError("Array fields may not specify hardcoded values.")
            if not self._delimited and not self._get_type().bounded:
                raise RuntimeError(
                    f"Unbounded element type ({self._type_string}) "
                    + "forbidden in non-delimited array."
                )
        else:
            if self._delimited:
                raise RuntimeError("Only arrays can be delimited.")

    def _validate_length_field(self):
        if self._length_field:
            if self._name is None:
                raise RuntimeError("Length fields must specify a name.")
            if self._hardcoded_value is not None:
                raise RuntimeError("Length fields may not specify hardcoded values.")
            field_type = self._get_type()
            if not isinstance(field_type, IntegerType):
                raise RuntimeError(
                    f"{field_type.name} is not a numeric type, "
                    + "so it is not allowed for a length field."
                )
        else:
            if self._offset != 0:
                raise RuntimeError("Only length fields can have an offset.")

    def _validate_unnamed_field(self):
        if self._name is not None:
            return

        if self._hardcoded_value is None:
            raise RuntimeError("Unnamed fields must specify a hardcoded field value.")

        if self._optional:
            raise RuntimeError("Unnamed fields may not be optional.")

    def _validate_hardcoded_value(self):
        if self._hardcoded_value is None:
            return

        field_type = self._get_type()

        if isinstance(field_type, StringType):
            length = try_parse_int(self._length_string)
            if length is not None and length != len(self._hardcoded_value):
                raise RuntimeError(
                    f"Expected length of {length} for hardcoded string value "
                    + f"'{self._hardcoded_value}'."
                )

        if not isinstance(field_type, BasicType):
            raise RuntimeError(
                f"Hardcoded field values are not allowed for {field_type.name} fields "
                + "(must be a basic type)."
            )

    def _validate_unique_name(self):
        if self._name is None:
            return

        if self._name in self._context.accessible_fields:
            raise RuntimeError(f"Cannot redefine {self._name} field.")

    def _validate_length_attribute(self):
        if self._length_string is None:
            return

        if (
            not self._length_string.isdigit()
            and self._length_string not in self._context.length_field_is_referenced_map
        ):
            raise RuntimeError(
                f'Length attribute "{self._length_string}" must be a numeric literal, '
                + 'or refer to a length field.'
            )

        is_already_referenced = self._context.length_field_is_referenced_map.get(
            self._length_string, False
        )

        if is_already_referenced:
            raise RuntimeError(
                f'Length field "{self._length_string}" must not be referenced by multiple fields.'
            )

    def generate_field(self):
        if self._name is None:
            return

        field_type = self._get_type()

        python_type_name = self._get_python_type_name(field_type)
        python_param_type_name = python_type_name

        if self._array_field:
            python_type_name = f"tuple[{python_type_name}, ...]"
            python_param_type_name = f"Iterable[{python_param_type_name}]"
            self._data.fields.add_import('annotations', '__future__')
            self._data.fields.add_import('Iterable', 'collections.abc')

        if self._optional:
            python_type_name = f"Optional[{python_type_name}]"
            python_param_type_name = f"Optional[{python_param_type_name}]"
            self._data.fields.add_import('Optional', 'typing')

        self._context.accessible_fields[self._name] = FieldData(
            self._name, field_type, self._offset, self._array_field
        )

        self._data.fields.add_line(f"_{self._name}: {python_type_name}")

        if isinstance(field_type, CustomType):
            self._data.fields.add_import_by_type(field_type)

        if self._length_field:
            self._context.length_field_is_referenced_map[self._name] = False
            return

        self._data.init_docstring_params.append(
            CodeBlock()
            .add(f'{self._name} ({python_param_type_name}): ')
            .add_code_block(self._generate_init_docstring())
        )

        self._data.add_method(
            CodeBlock()
            .add_line('@property')
            .add_line(f'def {self._name}(self) -> {python_type_name}:')
            .indent()
            .add_code_block(generate_docstring(self._comment))
            .add_line(f'return self._{self._name}')
            .unindent()
        )

        self._data.repr_fields.append(self._name)

        if self._hardcoded_value is None:
            expression = self._name
            if self._array_field:
                expression = f'tuple({expression})'
        elif isinstance(field_type, StringType):
            expression = f'"{self._hardcoded_value}"'
        else:
            expression = self._hardcoded_value

        init_param = CodeBlock().add(f'{self._name}: {python_param_type_name}')
        if self._optional:
            init_param.add(' = None')

        self._data.init_params.append(init_param)
        self._data.init_body.add_line(f'self._{self._name} = {expression}')

        if self._length_string in self._context.length_field_is_referenced_map:
            self._context.length_field_is_referenced_map[self._length_string] = True
            length_field_data = self._context.accessible_fields[self._length_string]
            self._data.init_body.add_line(
                f'self._{length_field_data.name} = len(self._{self._name})'
            )

    def generate_serialize(self):
        self._generate_serialize_missing_optional_guard()
        self._generate_serialize_none_not_allowed_error()
        self._generate_serialize_length_check()

        if self._array_field:
            array_size_expression = self._get_serialize_length_expression()
            if array_size_expression is None:
                array_size_expression = f"len(data._{self._name})"

            self._data.serialize.begin_control_flow(f"for i in range({array_size_expression})")

            if self._delimited and not self._trailing_delimiter:
                self._data.serialize.begin_control_flow("if i > 0")
                self._data.serialize.add_line("writer.add_byte(0xFF)")
                self._data.serialize.unindent()

        self._data.serialize.add_code_block(self._get_write_statement())

        if self._array_field:
            if self._delimited and self._trailing_delimiter:
                self._data.serialize.add_line("writer.add_byte(0xFF)")
            self._data.serialize.unindent()

        if self._optional:
            self._data.serialize.unindent()

    def generate_deserialize(self):
        if self._optional:
            python_type = f'Optional[{self._get_python_type_name(self._get_type())}]'
            self._data.deserialize.add_line(f'{self._name}: {python_type} = None')
            self._data.deserialize.begin_control_flow("if reader.remaining > 0")

        if self._array_field:
            self._generate_deserialize_array()
        else:
            self._data.deserialize.add_code_block(self._get_read_statement())

        if self._optional:
            self._data.deserialize.unindent()

        if self._name is not None and not self._length_field:
            self._data.deserialize_init_arguments.append(
                CodeBlock().add(f'{self._name}={self._name}')
            )

    def _generate_init_docstring(self):
        result = CodeBlock()

        if self._comment is not None:
            lines = map(str.strip, escape(self._comment, quote=False).split('\n'))
            for line in lines:
                if not result.empty:
                    result.add(' ')
                result.add(line)

        if self._length_string is not None:
            size_description = ""
            field_data = self._context.accessible_fields.get(self._length_string)
            if field_data:
                max_value = get_max_value_of(field_data.type_) + field_data.offset
                size_description = f"{max_value} or less"
            else:
                size_description = f'`{self._length_string}`'
                if self._padded:
                    size_description += " or less"
            if not result.empty:
                result.add(' ')
            result.add(f'(Length must be {size_description}.)')

        field_type = self._get_type()
        if isinstance(field_type, IntegerType):
            value_description = "Element value" if self._array_field else "Value"
            if not result.empty:
                result.add(' ')
            result.add(f'({value_description} range is 0-{get_max_value_of(field_type)}.)')

        return result

    def _generate_serialize_missing_optional_guard(self):
        if not self._optional:
            return

        if self._context.reached_optional_field:
            self._data.serialize.add_line(
                f"reached_missing_optional = reached_missing_optional or data._{self._name} is None"
            )
        else:
            self._data.serialize.add_line(f"reached_missing_optional = data._{self._name} is None")
        self._data.serialize.begin_control_flow("if not reached_missing_optional")

    def _generate_serialize_none_not_allowed_error(self):
        if self._optional or self._name is None or self._hardcoded_value is not None:
            return

        self._data.serialize.begin_control_flow(f"if data._{self._name} is None")
        self._data.serialize.add_line(f'raise SerializationError("{self._name} must be provided.")')
        self._data.serialize.unindent()
        self._data.serialize.add_import("SerializationError", "eolib.protocol.serialization_error")

    def _generate_serialize_length_check(self):
        if self._name is None:
            return

        length_expression = None
        field_data = self._context.accessible_fields.get(self._length_string)
        if field_data:
            length_expression = str(get_max_value_of(field_data.type_) + field_data.offset)
        else:
            length_expression = self._length_string

        if length_expression is None:
            return

        variable_size = self._padded or bool(field_data)
        length_check_operator = ">" if variable_size else "!="
        expected_length_description = (
            f"{length_expression} or less" if variable_size else f"exactly {length_expression}"
        )
        error_message = (
            "Expected length of "
            + self._name
            + " to be "
            + expected_length_description
            + (", got {len(data._" + self._name + ")}.")
        )

        self._data.serialize.begin_control_flow(
            f"if len(data._{self._name}) {length_check_operator} {length_expression}"
        )
        self._data.serialize.add_line(f'raise SerializationError(f"{error_message}")')
        self._data.serialize.unindent()
        self._data.serialize.add_import("SerializationError", "eolib.protocol.serialization_error")

    def _get_write_statement(self):
        real_type = self._get_type()
        type_ = real_type

        if isinstance(type_, HasUnderlyingType):
            type_ = type_.underlying_type

        value_expression = self._get_write_value_expression()
        if self._optional:
            value_expression = f'cast({self._get_python_type_name(real_type)}, {value_expression})'

        if isinstance(real_type, BoolType):
            value_expression = f"1 if {value_expression} else 0"

        if isinstance(real_type, EnumType):
            value_expression = f"int({value_expression})"

        offset_expression = FieldCodeGenerator._get_length_offset_expression(-self._offset)
        if offset_expression is not None:
            value_expression += offset_expression

        result = CodeBlock()

        if isinstance(type_, BasicType):
            length_expression = (
                None if self._array_field else self._get_serialize_length_expression()
            )
            write_statement = FieldCodeGenerator._get_write_statement_for_basic_type(
                type_, value_expression, length_expression, self._padded
            )
            result.add_line(write_statement)
        elif isinstance(type_, BlobType):
            result.add_line(f"writer.add_bytes({value_expression})")
        elif isinstance(type_, StructType):
            result.add_line(f"{type_.name}.serialize(writer, {value_expression})")
            result.add_import_by_type(type_)

        if not result:
            raise AssertionError("Unhandled Type")

        if self._optional:
            result.add_import("cast", "typing")

        return result

    def _get_write_value_expression(self):
        if self._name is None:
            type_ = self._get_type()
            if isinstance(type_, IntegerType):
                if self._hardcoded_value.isdigit():
                    return self._hardcoded_value
                raise RuntimeError(f'"{self._hardcoded_value}" is not a valid integer value.')
            elif isinstance(type_, BoolType):
                if self._hardcoded_value == "false":
                    return "0"
                elif self._hardcoded_value == "true":
                    return "1"
                raise RuntimeError(f'"{self._hardcoded_value}" is not a valid bool value.')
            elif isinstance(type_, StringType):
                return f'"{self._hardcoded_value}"'
            else:
                raise AssertionError("Unhandled BasicType")
        else:
            field_reference = f"data._{self._name}"
            if self._array_field:
                field_reference += "[i]"
            return field_reference

    @staticmethod
    def _get_write_statement_for_basic_type(type_, value_expression, length_expression, padded):
        if type_.name == "byte":
            return f"writer.add_byte({value_expression})"
        elif type_.name == "char":
            return f"writer.add_char({value_expression})"
        elif type_.name == "short":
            return f"writer.add_short({value_expression})"
        elif type_.name == "three":
            return f"writer.add_three({value_expression})"
        elif type_.name == "int":
            return f"writer.add_int({value_expression})"
        elif type_.name == "string":
            if length_expression is None:
                return f"writer.add_string({value_expression})"
            else:
                return f"writer.add_fixed_string({value_expression}, {length_expression}, {padded})"
        elif type_.name == "encoded_string":
            if length_expression is None:
                return f"writer.add_encoded_string({value_expression})"
            else:
                return (
                    "writer.add_fixed_encoded_string("
                    + f"{value_expression}, {length_expression}, {padded})"
                )
        else:
            raise AssertionError("Unhandled BasicType")

    def _generate_deserialize_array(self):
        array_length_expression = self._get_deserialize_length_expression()

        if array_length_expression is None and not self._delimited:
            element_size = self._get_type().fixed_size
            if element_size is not None:
                array_length_variable_name = f"{self._name}_length"
                self._data.deserialize.add_line(
                    f"{array_length_variable_name} = int(reader.remaining / {element_size})"
                )
                array_length_expression = array_length_variable_name

        self._data.deserialize.add_line(f"{self._name} = []")

        if array_length_expression is None:
            self._data.deserialize.begin_control_flow("while reader.remaining > 0")
        else:
            self._data.deserialize.begin_control_flow(f"for i in range({array_length_expression})")

        self._data.deserialize.add_code_block(self._get_read_statement())

        if self._delimited:
            needs_guard = not self._trailing_delimiter and array_length_expression is not None
            if needs_guard:
                self._data.deserialize.begin_control_flow(f"if i + 1 < {array_length_expression}")
            self._data.deserialize.add_line("reader.next_chunk()")
            if needs_guard:
                self._data.deserialize.unindent()

        self._data.deserialize.unindent()

    def _get_read_statement(self):
        real_type = self._get_type()
        type_ = real_type

        if isinstance(type_, HasUnderlyingType):
            type_ = type_.underlying_type

        statement = CodeBlock()

        if self._array_field:
            statement.add(f"{self._name}.append(")
        elif self._name is not None:
            statement.add(f"{self._name} = ")

        if isinstance(type_, BasicType):
            length_expression = (
                None if self._array_field else self._get_deserialize_length_expression()
            )
            read_basic_type = FieldCodeGenerator._get_read_statement_for_basic_type(
                type_, length_expression, self._padded
            )

            offset_expression = FieldCodeGenerator._get_length_offset_expression(self._offset)
            if offset_expression is not None:
                read_basic_type += offset_expression

            if isinstance(real_type, BoolType):
                statement.add(f"{read_basic_type} != 0")
            elif isinstance(real_type, EnumType):
                statement.add(f"{real_type.name}({read_basic_type})")
            else:
                statement.add(read_basic_type)
        elif isinstance(type_, BlobType):
            statement.add("reader.get_bytes(reader.remaining)")
        elif isinstance(type_, StructType):
            statement.add(f"{type_.name}.deserialize(reader)").add_import_by_type(type_)
        else:
            raise AssertionError("Unhandled Type")

        if self._array_field:
            statement.add(")")

        return statement.add("\n")

    def _check_field_accessible(self, field):
        field_data = self._context.accessible_fields.get(field)
        if not field_data:
            raise RuntimeError(f'Referenced {field} field is not accessible.')

    def _get_serialize_length_expression(self):
        expression = self._length_string
        if expression is not None and not expression.isdigit():
            self._check_field_accessible(expression)
            expression = f'data._{expression}'
        return expression

    def _get_deserialize_length_expression(self):
        expression = self._length_string
        if expression is not None and not expression.isdigit():
            self._check_field_accessible(expression)
        return expression

    @staticmethod
    def _get_length_offset_expression(offset):
        if offset != 0:
            return f" {'+' if offset > 0 else '-'} {abs(offset)}"
        return None

    @staticmethod
    def _get_read_statement_for_basic_type(type_, length_expression, padded):
        if type_.name == "byte":
            return "reader.get_byte()"
        elif type_.name == "char":
            return "reader.get_char()"
        elif type_.name == "short":
            return "reader.get_short()"
        elif type_.name == "three":
            return "reader.get_three()"
        elif type_.name == "int":
            return "reader.get_int()"
        elif type_.name == "string":
            if length_expression is None:
                return "reader.get_string()"
            else:
                return f"reader.get_fixed_string({length_expression}, {padded})"
        elif type_.name == "encoded_string":
            if length_expression is None:
                return "reader.get_encoded_string()"
            else:
                return f"reader.get_fixed_encoded_string({length_expression}, {padded})"
        else:
            raise AssertionError("Unhandled BasicType")

    def _get_type(self):
        return self._type_factory.get_type(self._type_string, self._get_type_length())

    def _get_type_length(self):
        if self._array_field:
            return Length.unspecified()

        if self._length_string is not None:
            return Length.from_string(self._length_string)

        return Length.unspecified()

    def _get_python_type_name(self, field_type):
        if isinstance(field_type, IntegerType):
            return "int"

        if isinstance(field_type, StringType):
            return "str"

        if isinstance(field_type, BoolType):
            return "bool"

        if isinstance(field_type, BlobType):
            return "bytes"

        if isinstance(field_type, CustomType):
            return field_type.name

        raise AssertionError("Unhandled Type")


class FieldCodeGeneratorBuilder:
    def __init__(self, type_factory, context, data):
        self._type_factory = type_factory
        self._context = context
        self._data = data
        self._name = None
        self._type = None
        self._length = None
        self._offset = 0
        self._padded = False
        self._optional = False
        self._hardcoded_value = None
        self._comment = None
        self._array_field = False
        self._length_field = False
        self._delimited = False
        self._trailing_delimiter = False

    def name(self, name):
        self._name = name
        return self

    def type(self, type_str):
        self._type = type_str
        return self

    def length(self, length):
        self._length = length
        return self

    def padded(self, padded):
        self._padded = padded
        return self

    def optional(self, optional):
        self._optional = optional
        return self

    def hardcoded_value(self, hardcoded_value):
        self._hardcoded_value = hardcoded_value
        return self

    def comment(self, comment):
        self._comment = comment
        return self

    def array_field(self, array_field):
        self._array_field = array_field
        return self

    def delimited(self, delimited):
        self._delimited = delimited
        return self

    def trailing_delimiter(self, trailing_delimiter):
        self._trailing_delimiter = trailing_delimiter
        return self

    def length_field(self, length_field):
        self._length_field = length_field
        return self

    def offset(self, offset):
        self._offset = offset
        return self

    def build(self):
        if self._type is None:
            raise ValueError("type must be provided")
        return FieldCodeGenerator(
            self._type_factory,
            self._context,
            self._data,
            self._name,
            self._type,
            self._length,
            self._padded,
            self._optional,
            self._hardcoded_value,
            self._comment,
            self._array_field,
            self._delimited,
            self._trailing_delimiter,
            self._length_field,
            self._offset,
        )


def get_max_value_of(integer_type):
    return 255 if integer_type.name == 'byte' else pow(253, integer_type.fixed_size) - 1



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ...coords import Coords
from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class ServerSettings:
    """
    Settings sent with WELCOME_REPLY packet
    """
    _byte_size: int = 0
    _jail_map: int
    _rescue_map: int
    _rescue_coords: Coords
    _spy_and_light_guide_flood_rate: int
    _guardian_flood_rate: int
    _game_master_flood_rate: int
    _high_game_master_flood_rate: int

    def __init__(self, *, jail_map: int, rescue_map: int, rescue_coords: Coords, spy_and_light_guide_flood_rate: int, guardian_flood_rate: int, game_master_flood_rate: int, high_game_master_flood_rate: int):
        """
        Create a new instance of ServerSettings.

        Args:
            jail_map (int): (Value range is 0-64008.)
            rescue_map (int): (Value range is 0-64008.)
            rescue_coords (Coords): 
            spy_and_light_guide_flood_rate (int): (Value range is 0-64008.)
            guardian_flood_rate (int): (Value range is 0-64008.)
            game_master_flood_rate (int): (Value range is 0-64008.)
            high_game_master_flood_rate (int): (Value range is 0-64008.)
        """
        self._jail_map = jail_map
        self._rescue_map = rescue_map
        self._rescue_coords = rescue_coords
        self._spy_and_light_guide_flood_rate = spy_and_light_guide_flood_rate
        self._guardian_flood_rate = guardian_flood_rate
        self._game_master_flood_rate = game_master_flood_rate
        self._high_game_master_flood_rate = high_game_master_flood_rate

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def jail_map(self) -> int:
        return self._jail_map

    @property
    def rescue_map(self) -> int:
        return self._rescue_map

    @property
    def rescue_coords(self) -> Coords:
        return self._rescue_coords

    @property
    def spy_and_light_guide_flood_rate(self) -> int:
        return self._spy_and_light_guide_flood_rate

    @property
    def guardian_flood_rate(self) -> int:
        return self._guardian_flood_rate

    @property
    def game_master_flood_rate(self) -> int:
        return self._game_master_flood_rate

    @property
    def high_game_master_flood_rate(self) -> int:
        return self._high_game_master_flood_rate

    @staticmethod
    def serialize(writer: EoWriter, data: "ServerSettings") -> None:
        """
        Serializes an instance of `ServerSettings` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (ServerSettings): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._jail_map is None:
                raise SerializationError("jail_map must be provided.")
            writer.add_short(data._jail_map)
            if data._rescue_map is None:
                raise SerializationError("rescue_map must be provided.")
            writer.add_short(data._rescue_map)
            if data._rescue_coords is None:
                raise SerializationError("rescue_coords must be provided.")
            Coords.serialize(writer, data._rescue_coords)
            if data._spy_and_light_guide_flood_rate is None:
                raise SerializationError("spy_and_light_guide_flood_rate must be provided.")
            writer.add_short(data._spy_and_light_guide_flood_rate)
            if data._guardian_flood_rate is None:
                raise SerializationError("guardian_flood_rate must be provided.")
            writer.add_short(data._guardian_flood_rate)
            if data._game_master_flood_rate is None:
                raise SerializationError("game_master_flood_rate must be provided.")
            writer.add_short(data._game_master_flood_rate)
            if data._high_game_master_flood_rate is None:
                raise SerializationError("high_game_master_flood_rate must be provided.")
            writer.add_short(data._high_game_master_flood_rate)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "ServerSettings":
        """
        Deserializes an instance of `ServerSettings` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            ServerSettings: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            jail_map = reader.get_short()
            rescue_map = reader.get_short()
            rescue_coords = Coords.deserialize(reader)
            spy_and_light_guide_flood_rate = reader.get_short()
            guardian_flood_rate = reader.get_short()
            game_master_flood_rate = reader.get_short()
            high_game_master_flood_rate = reader.get_short()
            result = ServerSettings(jail_map=jail_map, rescue_map=rescue_map, rescue_coords=rescue_coords, spy_and_light_guide_flood_rate=spy_and_light_guide_flood_rate, guardian_flood_rate=guardian_flood_rate, game_master_flood_rate=game_master_flood_rate, high_game_master_flood_rate=high_game_master_flood_rate)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"ServerSettings(byte_size={repr(self._byte_size)}, jail_map={repr(self._jail_map)}, rescue_map={repr(self._rescue_map)}, rescue_coords={repr(self._rescue_coords)}, spy_and_light_guide_flood_rate={repr(self._spy_and_light_guide_flood_rate)}, guardian_flood_rate={repr(self._guardian_flood_rate)}, game_master_flood_rate={repr(self._game_master_flood_rate)}, high_game_master_flood_rate={repr(self._high_game_master_flood_rate)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ..coords import Coords
from ...serialization_error import SerializationError
from ....data.eo_writer import EoWriter
from ....data.eo_reader import EoReader

class MapWarp:
    """
    Warp EMF entity
    """
    _byte_size: int = 0
    _destination_map: int
    _destination_coords: Coords
    _level_required: int
    _door: int

    def __init__(self, *, destination_map: int, destination_coords: Coords, level_required: int, door: int):
        """
        Create a new instance of MapWarp.

        Args:
            destination_map (int): (Value range is 0-64008.)
            destination_coords (Coords): 
            level_required (int): (Value range is 0-252.)
            door (int): (Value range is 0-64008.)
        """
        self._destination_map = destination_map
        self._destination_coords = destination_coords
        self._level_required = level_required
        self._door = door

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def destination_map(self) -> int:
        return self._destination_map

    @property
    def destination_coords(self) -> Coords:
        return self._destination_coords

    @property
    def level_required(self) -> int:
        return self._level_required

    @property
    def door(self) -> int:
        return self._door

    @staticmethod
    def serialize(writer: EoWriter, data: "MapWarp") -> None:
        """
        Serializes an instance of `MapWarp` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (MapWarp): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._destination_map is None:
                raise SerializationError("destination_map must be provided.")
            writer.add_short(data._destination_map)
            if data._destination_coords is None:
                raise SerializationError("destination_coords must be provided.")
            Coords.serialize(writer, data._destination_coords)
            if data._level_required is None:
                raise SerializationError("level_required must be provided.")
            writer.add_char(data._level_required)
            if data._door is None:
                raise SerializationError("door must be provided.")
            writer.add_short(data._door)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "MapWarp":
        """
        Deserializes an instance of `MapWarp` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            MapWarp: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            destination_map = reader.get_short()
            destination_coords = Coords.deserialize(reader)
            level_required = reader.get_char()
            door = reader.get_short()
            result = MapWarp(destination_map=destination_map, destination_coords=destination_coords, level_required=level_required, door=door)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"MapWarp(byte_size={repr(self._byte_size)}, destination_map={repr(self._destination_map)}, destination_coords={repr(self._destination_coords)}, level_required={repr(self._level_required)}, door={repr(self._door)})"



from abc import ABC, abstractproperty


class Type(ABC):
    @abstractproperty
    def name(self):
        pass

    @abstractproperty
    def fixed_size(self):
        pass

    @abstractproperty
    def bounded(self):
        pass



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

"""
Data structures generated from the [eo-protocol](https://github.com/cirras/eo-protocol/tree/master/xml/net/protocol.xml){target="_blank"} XML specification.

Warning:
  - This subpackage should not be directly imported. 
  - Instead, import [eolib.protocol.net][] (or the top-level `eolib`).
"""

from .weight import *
from .version import *
from .three_item import *
from .spell import *
from .quest_page import *
from .party_request_type import *
from .packet_family import *
from .packet_action import *
from .item import *
from .char_item import *



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class SkillStatRequirements:
    """
    Stat requirements to learn a skill from a skill master NPC
    """
    _byte_size: int = 0
    _str: int
    _wis: int
    _intl: int
    _agi: int
    _con: int
    _cha: int

    def __init__(self, *, str: int, wis: int, intl: int, agi: int, con: int, cha: int):
        """
        Create a new instance of SkillStatRequirements.

        Args:
            str (int): (Value range is 0-64008.)
            wis (int): (Value range is 0-64008.)
            intl (int): (Value range is 0-64008.)
            agi (int): (Value range is 0-64008.)
            con (int): (Value range is 0-64008.)
            cha (int): (Value range is 0-64008.)
        """
        self._str = str
        self._wis = wis
        self._intl = intl
        self._agi = agi
        self._con = con
        self._cha = cha

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def str(self) -> int:
        return self._str

    @property
    def wis(self) -> int:
        return self._wis

    @property
    def intl(self) -> int:
        return self._intl

    @property
    def agi(self) -> int:
        return self._agi

    @property
    def con(self) -> int:
        return self._con

    @property
    def cha(self) -> int:
        return self._cha

    @staticmethod
    def serialize(writer: EoWriter, data: "SkillStatRequirements") -> None:
        """
        Serializes an instance of `SkillStatRequirements` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (SkillStatRequirements): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._str is None:
                raise SerializationError("str must be provided.")
            writer.add_short(data._str)
            if data._wis is None:
                raise SerializationError("wis must be provided.")
            writer.add_short(data._wis)
            if data._intl is None:
                raise SerializationError("intl must be provided.")
            writer.add_short(data._intl)
            if data._agi is None:
                raise SerializationError("agi must be provided.")
            writer.add_short(data._agi)
            if data._con is None:
                raise SerializationError("con must be provided.")
            writer.add_short(data._con)
            if data._cha is None:
                raise SerializationError("cha must be provided.")
            writer.add_short(data._cha)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "SkillStatRequirements":
        """
        Deserializes an instance of `SkillStatRequirements` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            SkillStatRequirements: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            str = reader.get_short()
            wis = reader.get_short()
            intl = reader.get_short()
            agi = reader.get_short()
            con = reader.get_short()
            cha = reader.get_short()
            result = SkillStatRequirements(str=str, wis=wis, intl=intl, agi=agi, con=con, cha=cha)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"SkillStatRequirements(byte_size={repr(self._byte_size)}, str={repr(self._str)}, wis={repr(self._wis)}, intl={repr(self._intl)}, agi={repr(self._agi)}, con={repr(self._con)}, cha={repr(self._cha)})"



def interleave(data: bytearray) -> None:
    """
    Interleaves a sequence of bytes. When encrypting EO data, bytes are "woven" into each other.

    Used when encrypting packets and data files.

    Example:
        ```
        [0, 1, 2, 3, 4, 5] -> [0, 5, 1, 4, 2, 3]
        ```

    This is an in-place operation.

    Args:
        data (bytearray): The data to interleave.
    """
    buffer = bytearray(len(data))
    i = 0
    ii = 0

    while i < len(data):
        buffer[i] = data[ii]
        i += 2
        ii += 1

    i -= 1

    if len(data) % 2 != 0:
        i -= 2

    while i >= 0:
        buffer[i] = data[ii]
        i -= 2
        ii += 1

    data[:] = buffer


def deinterleave(data: bytearray) -> None:
    """
    Deinterleaves a sequence of bytes. This is the reverse of interleave.

    Used when decrypting packets and data files.

    Example:
        ```
        [0, 1, 2, 3, 4, 5] -> [0, 2, 4, 5, 3, 1]
        ```

    This is an in-place operation.

    Args:
        data (bytearray): The data to deinterleave.
    """
    buffer = bytearray(len(data))
    i = 0
    ii = 0

    while i < len(data):
        buffer[ii] = data[i]
        i += 2
        ii += 1

    i -= 1

    if len(data) % 2 != 0:
        i -= 2

    while i >= 0:
        buffer[ii] = data[i]
        i -= 2
        ii += 1

    data[:] = buffer


def flip_msb(data: bytearray) -> None:
    """
    Flips the most significant bits of each byte in a sequence of bytes.
    (Values 0 and 128 are not flipped.)

    Used when encrypting and decrypting packets.

    Example:
        ```
        [0, 1, 127, 128, 129, 254, 255] -> [0, 129, 255, 128, 1, 126, 127]
        ```

    This is an in-place operation.

    Args:
        data (bytearray): The data to flip most significant bits on.
    """
    for i in range(len(data)):
        if data[i] & 0x7F != 0:
            data[i] = data[i] ^ 0x80


def swap_multiples(data: bytearray, multiple: int) -> None:
    """
    Swaps the order of contiguous bytes in a sequence of bytes that are divisible by a given
    multiple value.

    Used when encrypting and decrypting packets and data files.

    Example:
        ```
        multiple = 3
        [10, 21, 27] -> [10, 27, 21]
        ```

    This is an in-place operation.

    Args:
        data (bytearray): The data to swap bytes in.
        multiple (int): The multiple value.
    Raises:
        ValueError: If multiple is less than 1.
    """
    if multiple < 0:
        raise ValueError("multiple must be a positive number")

    if multiple == 0:
        return

    sequence_length = 0

    for i in range(len(data) + 1):
        if i != len(data) and data[i] % multiple == 0:
            sequence_length += 1
        else:
            if sequence_length > 1:
                for ii in range(sequence_length // 2):
                    b = data[i - sequence_length + ii]
                    data[i - sequence_length + ii] = data[i - ii - 1]
                    data[i - ii - 1] = b

            sequence_length = 0



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from .shop_sold_item import ShopSoldItem
from ..weight import Weight
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class ShopSellServerPacket(Packet):
    """
    Response to selling an item to a shop
    """
    _byte_size: int = 0
    _sold_item: ShopSoldItem
    _gold_amount: int
    _weight: Weight

    def __init__(self, *, sold_item: ShopSoldItem, gold_amount: int, weight: Weight):
        """
        Create a new instance of ShopSellServerPacket.

        Args:
            sold_item (ShopSoldItem): 
            gold_amount (int): (Value range is 0-4097152080.)
            weight (Weight): 
        """
        self._sold_item = sold_item
        self._gold_amount = gold_amount
        self._weight = weight

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def sold_item(self) -> ShopSoldItem:
        return self._sold_item

    @property
    def gold_amount(self) -> int:
        return self._gold_amount

    @property
    def weight(self) -> Weight:
        return self._weight

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Shop

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Sell

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        ShopSellServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "ShopSellServerPacket") -> None:
        """
        Serializes an instance of `ShopSellServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (ShopSellServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._sold_item is None:
                raise SerializationError("sold_item must be provided.")
            ShopSoldItem.serialize(writer, data._sold_item)
            if data._gold_amount is None:
                raise SerializationError("gold_amount must be provided.")
            writer.add_int(data._gold_amount)
            if data._weight is None:
                raise SerializationError("weight must be provided.")
            Weight.serialize(writer, data._weight)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "ShopSellServerPacket":
        """
        Deserializes an instance of `ShopSellServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            ShopSellServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            sold_item = ShopSoldItem.deserialize(reader)
            gold_amount = reader.get_int()
            weight = Weight.deserialize(reader)
            result = ShopSellServerPacket(sold_item=sold_item, gold_amount=gold_amount, weight=weight)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"ShopSellServerPacket(byte_size={repr(self._byte_size)}, sold_item={repr(self._sold_item)}, gold_amount={repr(self._gold_amount)}, weight={repr(self._weight)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from typing import Union
from .marriage_reply import MarriageReply
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class MarriageReplyServerPacket(Packet):
    """
    Reply to client Marriage-family packets
    """
    _byte_size: int = 0
    _reply_code: MarriageReply
    _reply_code_data: 'MarriageReplyServerPacket.ReplyCodeData'

    def __init__(self, *, reply_code: MarriageReply, reply_code_data: 'MarriageReplyServerPacket.ReplyCodeData' = None):
        """
        Create a new instance of MarriageReplyServerPacket.

        Args:
            reply_code (MarriageReply): 
            reply_code_data (MarriageReplyServerPacket.ReplyCodeData): Data associated with the `reply_code` field.
        """
        self._reply_code = reply_code
        self._reply_code_data = reply_code_data

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def reply_code(self) -> MarriageReply:
        return self._reply_code

    @property
    def reply_code_data(self) -> 'MarriageReplyServerPacket.ReplyCodeData':
        """
        MarriageReplyServerPacket.ReplyCodeData: Data associated with the `reply_code` field.
        """
        return self._reply_code_data

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Marriage

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Reply

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        MarriageReplyServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "MarriageReplyServerPacket") -> None:
        """
        Serializes an instance of `MarriageReplyServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (MarriageReplyServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._reply_code is None:
                raise SerializationError("reply_code must be provided.")
            writer.add_short(int(data._reply_code))
            if data._reply_code == MarriageReply.Success:
                if not isinstance(data._reply_code_data, MarriageReplyServerPacket.ReplyCodeDataSuccess):
                    raise SerializationError("Expected reply_code_data to be type MarriageReplyServerPacket.ReplyCodeDataSuccess for reply_code " + MarriageReply(data._reply_code).name + ".")
                MarriageReplyServerPacket.ReplyCodeDataSuccess.serialize(writer, data._reply_code_data)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "MarriageReplyServerPacket":
        """
        Deserializes an instance of `MarriageReplyServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            MarriageReplyServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reply_code = MarriageReply(reader.get_short())
            reply_code_data: MarriageReplyServerPacket.ReplyCodeData = None
            if reply_code == MarriageReply.Success:
                reply_code_data = MarriageReplyServerPacket.ReplyCodeDataSuccess.deserialize(reader)
            result = MarriageReplyServerPacket(reply_code=reply_code, reply_code_data=reply_code_data)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"MarriageReplyServerPacket(byte_size={repr(self._byte_size)}, reply_code={repr(self._reply_code)}, reply_code_data={repr(self._reply_code_data)})"

    ReplyCodeData = Union['MarriageReplyServerPacket.ReplyCodeDataSuccess', None]
    """
    Data associated with different values of the `reply_code` field.
    """

    class ReplyCodeDataSuccess:
        """
        Data associated with reply_code value MarriageReply.Success
        """
        _byte_size: int = 0
        _gold_amount: int

        def __init__(self, *, gold_amount: int):
            """
            Create a new instance of MarriageReplyServerPacket.ReplyCodeDataSuccess.

            Args:
                gold_amount (int): (Value range is 0-4097152080.)
            """
            self._gold_amount = gold_amount

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def gold_amount(self) -> int:
            return self._gold_amount

        @staticmethod
        def serialize(writer: EoWriter, data: "MarriageReplyServerPacket.ReplyCodeDataSuccess") -> None:
            """
            Serializes an instance of `MarriageReplyServerPacket.ReplyCodeDataSuccess` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (MarriageReplyServerPacket.ReplyCodeDataSuccess): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._gold_amount is None:
                    raise SerializationError("gold_amount must be provided.")
                writer.add_int(data._gold_amount)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "MarriageReplyServerPacket.ReplyCodeDataSuccess":
            """
            Deserializes an instance of `MarriageReplyServerPacket.ReplyCodeDataSuccess` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                MarriageReplyServerPacket.ReplyCodeDataSuccess: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                gold_amount = reader.get_int()
                result = MarriageReplyServerPacket.ReplyCodeDataSuccess(gold_amount=gold_amount)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"MarriageReplyServerPacket.ReplyCodeDataSuccess(byte_size={repr(self._byte_size)}, gold_amount={repr(self._gold_amount)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class DialogEntryType(IntEnum, metaclass=ProtocolEnumMeta):
    """
    The type of an entry in a quest dialog
    """
    Text = 1
    Link = 2



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class InnUnsubscribeReply(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Reply code sent with CITIZEN_REMOVE packet.
    Indicates the result of trying to give up citizenship to a town.
    """
    NotCitizen = 0
    Unsubscribed = 1



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

"""
Data structures generated from the [eo-protocol](https://github.com/cirras/eo-protocol/tree/master/xml/net/server/protocol.xml){target="_blank"} XML specification.

Warning:
  - This subpackage should not be directly imported. 
  - Instead, import [eolib.protocol.net.server][] (or the top-level `eolib`).
"""

from .welcome_reply_server_packet import *
from .welcome_pong_server_packet import *
from .welcome_ping_server_packet import *
from .welcome_net244_server_packet import *
from .welcome_net243_server_packet import *
from .welcome_net242_server_packet import *
from .welcome_code import *
from .warp_type import *
from .warp_request_server_packet import *
from .warp_player_server_packet import *
from .warp_effect import *
from .warp_create_server_packet import *
from .warp_agree_server_packet import *
from .walk_reply_server_packet import *
from .walk_player_server_packet import *
from .walk_open_server_packet import *
from .walk_close_server_packet import *
from .trade_use_server_packet import *
from .trade_spec_server_packet import *
from .trade_request_server_packet import *
from .trade_reply_server_packet import *
from .trade_open_server_packet import *
from .trade_item_data import *
from .trade_close_server_packet import *
from .trade_agree_server_packet import *
from .trade_admin_server_packet import *
from .tile_effect import *
from .talk_tell_server_packet import *
from .talk_spec_server_packet import *
from .talk_server_server_packet import *
from .talk_request_server_packet import *
from .talk_reply_server_packet import *
from .talk_reply import *
from .talk_player_server_packet import *
from .talk_open_server_packet import *
from .talk_msg_server_packet import *
from .talk_list_server_packet import *
from .talk_announce_server_packet import *
from .talk_admin_server_packet import *
from .stat_skill_take_server_packet import *
from .stat_skill_reply_server_packet import *
from .stat_skill_remove_server_packet import *
from .stat_skill_player_server_packet import *
from .stat_skill_open_server_packet import *
from .stat_skill_junk_server_packet import *
from .stat_skill_accept_server_packet import *
from .spell_target_self_server_packet import *
from .spell_target_other_server_packet import *
from .spell_target_group_server_packet import *
from .spell_request_server_packet import *
from .spell_reply_server_packet import *
from .spell_player_server_packet import *
from .spell_error_server_packet import *
from .skill_stat_requirements import *
from .skill_master_reply import *
from .skill_learn import *
from .sit_state import *
from .sit_reply_server_packet import *
from .sit_remove_server_packet import *
from .sit_player_server_packet import *
from .sit_close_server_packet import *
from .shop_trade_item import *
from .shop_sold_item import *
from .shop_sell_server_packet import *
from .shop_open_server_packet import *
from .shop_create_server_packet import *
from .shop_craft_item import *
from .shop_buy_server_packet import *
from .server_settings import *
from .refresh_reply_server_packet import *
from .recover_target_group_server_packet import *
from .recover_reply_server_packet import *
from .recover_player_server_packet import *
from .recover_list_server_packet import *
from .recover_agree_server_packet import *
from .range_reply_server_packet import *
from .quest_requirement_icon import *
from .quest_report_server_packet import *
from .quest_progress_entry import *
from .quest_list_server_packet import *
from .quest_dialog_server_packet import *
from .pub_file import *
from .priest_request_server_packet import *
from .priest_reply_server_packet import *
from .priest_reply import *
from .priest_open_server_packet import *
from .players_reply_server_packet import *
from .players_remove_server_packet import *
from .players_pong_server_packet import *
from .players_ping_server_packet import *
from .players_net242_server_packet import *
from .players_list_server_packet import *
from .players_list_friends import *
from .players_list import *
from .players_agree_server_packet import *
from .player_killed_state import *
from .player_effect import *
from .party_target_group_server_packet import *
from .party_request_server_packet import *
from .party_reply_server_packet import *
from .party_reply_code import *
from .party_remove_server_packet import *
from .party_member import *
from .party_list_server_packet import *
from .party_exp_share import *
from .party_create_server_packet import *
from .party_close_server_packet import *
from .party_agree_server_packet import *
from .party_add_server_packet import *
from .paperdoll_reply_server_packet import *
from .paperdoll_remove_server_packet import *
from .paperdoll_ping_server_packet import *
from .paperdoll_agree_server_packet import *
from .online_player import *
from .npc_update_position import *
from .npc_update_chat import *
from .npc_update_attack import *
from .npc_spec_server_packet import *
from .npc_reply_server_packet import *
from .npc_player_server_packet import *
from .npc_map_info import *
from .npc_killed_data import *
from .npc_kill_steal_protection_state import *
from .npc_junk_server_packet import *
from .npc_dialog_server_packet import *
from .npc_agree_server_packet import *
from .npc_accept_server_packet import *
from .nearby_info import *
from .music_player_server_packet import *
from .message_pong_server_packet import *
from .message_open_server_packet import *
from .message_close_server_packet import *
from .message_accept_server_packet import *
from .marriage_reply_server_packet import *
from .marriage_reply import *
from .marriage_open_server_packet import *
from .map_file import *
from .map_effect import *
from .map_drain_damage_other import *
from .map_damage_type import *
from .login_reply_server_packet import *
from .login_reply import *
from .login_message_code import *
from .locker_spec_server_packet import *
from .locker_reply_server_packet import *
from .locker_open_server_packet import *
from .locker_get_server_packet import *
from .locker_buy_server_packet import *
from .level_up_stats import *
from .jukebox_use_server_packet import *
from .jukebox_reply_server_packet import *
from .jukebox_player_server_packet import *
from .jukebox_open_server_packet import *
from .jukebox_msg_server_packet import *
from .jukebox_agree_server_packet import *
from .item_spec_server_packet import *
from .item_reply_server_packet import *
from .item_remove_server_packet import *
from .item_obtain_server_packet import *
from .item_map_info import *
from .item_kick_server_packet import *
from .item_junk_server_packet import *
from .item_get_server_packet import *
from .item_drop_server_packet import *
from .item_agree_server_packet import *
from .item_add_server_packet import *
from .item_accept_server_packet import *
from .inn_unsubscribe_reply import *
from .init_reply import *
from .init_init_server_packet import *
from .init_ban_type import *
from .guild_tell_server_packet import *
from .guild_take_server_packet import *
from .guild_staff import *
from .guild_sell_server_packet import *
from .guild_request_server_packet import *
from .guild_report_server_packet import *
from .guild_reply_server_packet import *
from .guild_reply import *
from .guild_rank_server_packet import *
from .guild_open_server_packet import *
from .guild_member import *
from .guild_kick_server_packet import *
from .guild_create_server_packet import *
from .guild_buy_server_packet import *
from .guild_agree_server_packet import *
from .guild_accept_server_packet import *
from .group_heal_target_player import *
from .global_backfill_message import *
from .face_player_server_packet import *
from .equipment_welcome import *
from .equipment_paperdoll import *
from .equipment_map_info import *
from .equipment_character_select import *
from .equipment_change import *
from .emote_player_server_packet import *
from .effect_use_server_packet import *
from .effect_target_other_server_packet import *
from .effect_spec_server_packet import *
from .effect_report_server_packet import *
from .effect_player_server_packet import *
from .effect_agree_server_packet import *
from .effect_admin_server_packet import *
from .door_open_server_packet import *
from .door_close_server_packet import *
from .dialog_quest_entry import *
from .dialog_entry_type import *
from .dialog_entry import *
from .connection_player_server_packet import *
from .citizen_request_server_packet import *
from .citizen_reply_server_packet import *
from .citizen_remove_server_packet import *
from .citizen_open_server_packet import *
from .citizen_accept_server_packet import *
from .chest_spec_server_packet import *
from .chest_reply_server_packet import *
from .chest_open_server_packet import *
from .chest_get_server_packet import *
from .chest_close_server_packet import *
from .chest_agree_server_packet import *
from .character_stats_welcome import *
from .character_stats_update import *
from .character_stats_reset import *
from .character_stats_info_lookup import *
from .character_stats_equipment_change import *
from .character_selection_list_entry import *
from .character_secondary_stats_info_lookup import *
from .character_secondary_stats import *
from .character_reply_server_packet import *
from .character_reply import *
from .character_player_server_packet import *
from .character_map_info import *
from .character_icon import *
from .character_elemental_stats import *
from .character_details import *
from .character_base_stats_welcome import *
from .character_base_stats import *
from .chair_reply_server_packet import *
from .chair_remove_server_packet import *
from .chair_player_server_packet import *
from .chair_close_server_packet import *
from .cast_spec_server_packet import *
from .cast_reply_server_packet import *
from .cast_accept_server_packet import *
from .book_reply_server_packet import *
from .board_post_listing import *
from .board_player_server_packet import *
from .board_open_server_packet import *
from .big_coords import *
from .barber_open_server_packet import *
from .barber_agree_server_packet import *
from .bank_reply_server_packet import *
from .bank_open_server_packet import *
from .avatar_reply_server_packet import *
from .avatar_remove_server_packet import *
from .avatar_change_type import *
from .avatar_change import *
from .avatar_agree_server_packet import *
from .avatar_admin_server_packet import *
from .attack_player_server_packet import *
from .attack_error_server_packet import *
from .arena_use_server_packet import *
from .arena_spec_server_packet import *
from .arena_drop_server_packet import *
from .arena_accept_server_packet import *
from .admin_message_type import *
from .admin_interact_tell_server_packet import *
from .admin_interact_reply_server_packet import *
from .admin_interact_remove_server_packet import *
from .admin_interact_list_server_packet import *
from .admin_interact_agree_server_packet import *
from .account_reply_server_packet import *
from .account_reply import *



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class ShopSoldItem:
    """
    A sold item when selling an item to a shop
    """
    _byte_size: int = 0
    _amount: int
    _id: int

    def __init__(self, *, amount: int, id: int):
        """
        Create a new instance of ShopSoldItem.

        Args:
            amount (int): (Value range is 0-4097152080.)
            id (int): (Value range is 0-64008.)
        """
        self._amount = amount
        self._id = id

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def amount(self) -> int:
        return self._amount

    @property
    def id(self) -> int:
        return self._id

    @staticmethod
    def serialize(writer: EoWriter, data: "ShopSoldItem") -> None:
        """
        Serializes an instance of `ShopSoldItem` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (ShopSoldItem): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._amount is None:
                raise SerializationError("amount must be provided.")
            writer.add_int(data._amount)
            if data._id is None:
                raise SerializationError("id must be provided.")
            writer.add_short(data._id)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "ShopSoldItem":
        """
        Deserializes an instance of `ShopSoldItem` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            ShopSoldItem: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            amount = reader.get_int()
            id = reader.get_short()
            result = ShopSoldItem(amount=amount, id=id)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"ShopSoldItem(byte_size={repr(self._byte_size)}, amount={repr(self._amount)}, id={repr(self._id)})"
